import numpy as np
import warnings
from ..utils import DESC_ENCODING, TARG_ENCODING, MISS_ENCODING


def base_selection_algorithm(metadata, nb_targets=1, nb_iterations=1, random_state=997):
    m_codes = random_selection_algorithm(
        metadata,
        nb_targets=nb_targets,
        nb_iterations=nb_iterations,
        fraction_missing=0.0,
        random_state=random_state,
    )
    return m_codes


def random_selection_algorithm(
    metadata, nb_targets=1, nb_iterations=1, fraction_missing=0.2, random_state=997
):
    if isinstance(fraction_missing, list):
        codes = []
        for f in fraction_missing:
            codes.append(
                random_selection_algorithm(
                    metadata,
                    nb_targets=nb_targets,
                    nb_iterations=nb_iterations,
                    fraction_missing=f,
                    random_state=random_state,
                )
            )
        m_codes = np.vstack(codes)
        return m_codes
    
    else:
        # Init
        np.random.seed(random_state)
        nb_attributes = metadata["n_attributes"]
        nb_targets = _set_nb_targets(nb_targets, nb_attributes)

        if fraction_missing in {'sqrt'}:
            fraction_missing = 1-np.sqrt(nb_attributes)/nb_attributes
        elif fraction_missing in {"log2"}:
            fraction_missing = 1-np.log2(nb_attributes)/nb_attributes

        codes = []
        for attribute_kind in {"nominal_attributes", "numeric_attributes"}:
            potential_targets = np.array(list(metadata[attribute_kind]))

            if potential_targets.shape[0] > 0:
                for iterations in range(nb_iterations):
                    m_codes = _single_iteration_random_selection(
                        nb_attributes, nb_targets, fraction_missing, potential_targets
                    )
                    codes.append(m_codes)

        m_codes = np.vstack(codes)

        m_codes = _ensure_desc_atts(m_codes)
        return m_codes.astype(np.int8)


def _single_iteration_random_selection(
    nb_attributes, nb_targets, fraction_missing, potential_targets
):
    nb_models, deficit = _nb_models_and_deficit(nb_targets, potential_targets)

    # Init
    target_sets, nb_models = _target_sets(potential_targets, nb_targets, nb_models, deficit)

    m_codes = _init(nb_models, nb_attributes)

    m_codes = _set_targets(m_codes, target_sets)
    m_codes = _set_missing(m_codes, fraction_missing)
    m_codes
    return m_codes


# Helpers
def _set_missing(m_codes, fraction=0.2):
    random = np.random.rand(*m_codes.shape)

    noise = np.where(m_codes == 0, random, m_codes)
    missing = np.where(noise < fraction, -1, noise)

    res = np.floor(missing)

    return res


def _ensure_desc_atts(m_codes):
    """
    If there are no input attributes in a code, we flip one missing attribute at random.
    """
    for row in m_codes:
        if 0 not in np.unique(row):
            idx_of_minus_ones = np.where(row == -1)[0]
            idx_to_change_to_zero = np.random.choice(idx_of_minus_ones)
            row[idx_to_change_to_zero] = 0

    return m_codes


def _set_nb_targets(nb_targets, nb_atts):
    if (nb_targets > 0) & (nb_targets < 1):
        nb_out_atts = int(np.ceil(nb_targets * nb_atts))
    elif (nb_targets >= 1) & (nb_targets < nb_atts):
        nb_out_atts = int(nb_targets)
    else:
        msg = """
        Impossible number of output attributes per model: {}
        This means the value of `nb_targets` was set incorrectly.
        Re-adjusted to default=1; one model per attribute.
        """.format(
            nb_targets
        )
        warnings.warn(msg)
        nb_out_atts = 1

    return nb_out_atts


def _nb_models_and_deficit(nb_targets, potential_targets):

    nb_potential_targets = potential_targets.shape[0]

    nb_models_with_regular_nb_targets = nb_potential_targets // nb_targets
    nb_leftover_targets = nb_potential_targets % nb_targets

    if nb_leftover_targets:
        nb_models = nb_models_with_regular_nb_targets 
        deficit = nb_leftover_targets
    else:
        nb_models = nb_models_with_regular_nb_targets
        deficit = 0

    return nb_models, deficit


def _init(nb_models, nb_attributes):
    return np.zeros((nb_models, nb_attributes), dtype=np.int8)


def _target_sets(potential_targets, nb_targets, nb_models, deficit):

    nb_targets = min(len(potential_targets), nb_targets)
    
    np.random.shuffle(potential_targets)

    choices = potential_targets[deficit:]
    result = np.random.choice(choices, replace=False, size=(nb_models, nb_targets))

    if deficit:
        choices = potential_targets[:nb_targets] # This includes all the ones you left out!
        extra = np.random.choice(choices, replace=False, size=(1, nb_targets))
        result = np.vstack([result, extra])
        nb_models += 1

    return result, nb_models


def _set_targets(m_codes, target_sets):

    row_idx = np.arange(m_codes.shape[0]).reshape(-1, 1)
    col_idx = target_sets

    m_codes[row_idx, col_idx] = TARG_ENCODING
    return m_codes




from functools import partial

import networkx as nx
import numpy as np

from ..composition import o
from ..utils.inference_tools import (
    _dummy_array,
    _map_classes,
    _pad_proba,
    _select_nominal,
    _select_numeric,
)


# Main algorithm
def inference_algorithm(g, m_list, i_list, c_list, data, nominal_ids):
    """Add inference information to graph g
    
    Arguments:
        g {[type]} -- [description]
        m_list {[type]} -- [description]
        i_list {[type]} -- [description]
        c_list {[type]} -- [description]
        data {[type]} -- [description]
        nominal_ids {[type]} -- [description]
    
    Raises:
        ValueError: [description]
    
    Returns:
        [type] -- [description]
    """

    _data_node = lambda k, n: k == "D"
    _model_node = lambda k, n: k == "M"
    _imputation_node = lambda k, n: k == "I"
    _composite_node = lambda k, n: k == "C"

    nb_rows, _ = data.shape

    g_desc_ids = list(g.desc_ids)

    if data is not None:
        g.data = data[:, g_desc_ids]
    else:
        g.data = None

    for n in g.nodes():
        if _data_node(*n):
            in_degree = g.in_degree(n)
            if in_degree == 0:
                input_data_node(g, n, g_desc_ids)
            elif in_degree > 0:
                if n[1] in nominal_ids:
                    nominal_data_node(g, n, m_list, c_list)
                else:
                    numeric_data_node(g, n, m_list, c_list)
        elif _model_node(*n):
            model_node(g, n, m_list)
        elif _imputation_node(*n):
            imputation_node(g, n, i_list, nb_rows)
        elif _composite_node(*n):
            composite_node(g, n, c_list)
        else:
            raise ValueError("Did not recognize node kind of {}".format(n))

    return


# Specific Nodes
def input_data_node(g, node, g_desc_ids):
    def f(rel_idx):
        f1 = _select_numeric(rel_idx)
        return f1(g.data)

    # New
    g.nodes[node]["inputs"] = g_desc_ids.index(node[1])
    g.nodes[node]["compute"] = f

    return


def imputation_node(g, node, i_list, nb_rows):

    # Build function
    def f(n):
        return i_list[node[1]].transform(_dummy_array(n)).ravel()

    # New
    g.nodes[node]["inputs"] = nb_rows
    g.nodes[node]["compute"] = f
    return

"""
def single_data_node(g, node, m_list, c_list):
    # New
    parents = _numeric_parents(g, m_list, c_list, node)

    def f(parents):
        collector = _numeric_inputs(g, parents)
        return collector.pop()

    g.nodes[node]["inputs"] = parents
    g.nodes[node]["compute"] = f

    return
"""

def numeric_data_node(g, node, m_list, c_list):
    parents = _numeric_parents(g, m_list, c_list, node)

    def f(parents):
        collector = _numeric_inputs(g, parents)
        return np.mean(collector, axis=0)

    g.nodes[node]["inputs"] = parents
    g.nodes[node]["compute"] = f
    return


def nominal_data_node(g, node, m_list, c_list):
    # New
    parents = _nominal_parents(g, m_list, c_list, node)

    classes = np.unique(np.hstack([c for _, c, _ in parents]))

    def vote(X):
        return classes.take(np.argmax(X, axis=1), axis=0)

    def F(parents):
        collector = _nominal_inputs(g, parents, classes)
        return np.sum(collector, axis=0)

    def F2(parents):
        return vote(F(parents))

    g.nodes[node]["classes"] = classes
    g.nodes[node]["inputs"] = parents
    g.nodes[node]["compute_proba"] = F
    g.nodes[node]["compute"] = F2

    return


def model_node(g, node, m_list):

    # New
    parents = _model_parents(g, node)

    def f(parents):
        X = _model_inputs(g, parents)
        return m_list[node[1]].predict(X)

    g.nodes[node]["inputs"] = parents
    g.nodes[node]["compute"] = f

    if hasattr(m_list[node[1]], "predict_proba"):

        def f2(parents):
            X = _model_inputs(g, parents)
            return m_list[node[1]].predict_proba(X)

        g.nodes[node]["compute_proba"] = f2

    return


def composite_node(g, node, c_list):
    return model_node(g, node, c_list)


# Helpers - Function
def compute(g, node, proba=False):

    result = "result"
    compute = "compute"

    if proba:
        result += "_proba"
        compute += "_proba"

    r = g.nodes[node].get(result, None)
    if r is None:
        i = g.nodes[node].get("inputs")
        f = g.nodes[node].get(compute)
        g.nodes[node][result] = f(i)
        return g.nodes[node][result]
    else:
        return r


def _nominal_inputs(g, parents, classes):
    collector = [
        _select_nominal(idx)(compute(g, n, proba=True))
        if len(c) == len(classes)
        else _pad_proba(c, classes)(_select_nominal(idx)(compute(g, n, proba=True)))
        for idx, c, n in parents
    ]
    return collector


def _numeric_inputs(g, parents):
    collector = [_select_numeric(idx)(compute(g, n)) for idx, n in parents]
    return collector


def _model_inputs(g, parents):
    collector = [compute(g, n) for n in parents]
    collector = np.stack(collector, axis=1)
    return collector


def _numeric_parents(g, m_list, c_list, node):

    parents = [
        (rel_idx(p_idx, node[1], k, m_list, c_list), (k, p_idx))
        for k, p_idx in g.predecessors(node)
    ]

    return parents


def _nominal_parents(g, m_list, c_list, node):
    parents = [
        (
            rel_idx(p_idx, node[1], k, m_list, c_list),
            classes(
                p_idx, rel_idx(p_idx, node[1], k, m_list, c_list), k, m_list, c_list
            ),
            (k, p_idx),
        )
        for k, p_idx in g.predecessors(node)
    ]

    return parents


def _model_parents(g, node):
    idxs = {p_idx: (m, p_idx) for m, p_idx in g.predecessors(node)}

    parents = [n for k, n in sorted(idxs.items())]

    return parents


def rel_idx(p_idx, n_idx, k, m_list, c_list):
    if k == "M":
        return m_list[p_idx].targ_ids.index(n_idx)
    elif k == "C":
        return c_list[p_idx].targ_ids.index(n_idx)
    else:
        return 0


def classes(p_idx, r_idx, k, m_list, c_list):
    if k == "M":
        return m_list[p_idx].classes_[r_idx]
    elif k == "C":
        return c_list[p_idx].classes_[r_idx]




import os
from os.path import dirname

import numpy as np
import pandas as pd
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split

RANDOM_STATE = 42


def data_directory():
    test_dir = dirname(__file__)
    data_dir = os.path.join(test_dir, "data")  # Slight HARDCODE!
    return data_dir


def filename_data(dataset="iris", kind="train", separator="-", extension="csv"):

    data_dir = data_directory()
    filename = separator.join([dataset, kind]) + ".{}".format(extension)
    return os.path.join(data_dir, filename)


def _detect_nominal(df):
    nominal_columns = set(df.select_dtypes(exclude=["float"]).columns)

    nominal_ids = {i for i, c in enumerate(df.columns) if c in nominal_columns}

    return nominal_ids


def load_iris(df=False, nominal_ids=True):

    fn_train = filename_data(dataset="iris", kind="train")
    fn_test = filename_data(dataset="iris", kind="test")

    df_train = pd.read_csv(fn_train, header=None)
    df_test = pd.read_csv(fn_test, header=None)

    result = []
    result.append(df_train if df else df_train.values)
    result.append(df_test if df else df_test.values)

    if nominal_ids:
        result.append(_detect_nominal(df_train))

    return result


def default_dataset(
    random_state=RANDOM_STATE,
    n_samples=10 ** 3,
    n_features=7,
    n_redundant=0,
    n_repeated=0,
    test_size=0.2,
    n_clusters_per_class=2,
    df=False,
    **kwargs
):
    """
    Generate a dataset to be used in tests.

    Returns:

    """
    X, y = make_classification(
        n_samples=n_samples,
        n_features=n_features,
        n_repeated=n_repeated,
        n_redundant=n_redundant,
        n_clusters_per_class=n_clusters_per_class,
        random_state=random_state,
        **kwargs
    )

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=test_size, random_state=random_state
    )

    if df:
        train = pd.DataFrame(X_train)
        train = train.assign(y=y_train)

        test = pd.DataFrame(X_test)
        test = test.assign(y=y_test)

        return train, test
    else:
        return np.c_[X_train, y_train], np.c_[X_test, y_test]



from sklearn.impute import SimpleImputer
import numpy as np


def nan_imputation(X, nominal_attributes):
    # Init
    n_rows, n_cols = X.shape
    i_list = []

    # Make imputers
    for c in range(n_cols):
        i_config = dict(missing_values=np.nan, strategy="constant", fill_value=np.nan)

        # Initialize imputer
        i = SimpleImputer(**i_config)
        i.fit(X[:, [c]])

        if c in nominal_attributes:
            i.out_kind = "nominal"
        else:
            i.out_kind = "numeric"

        i_list.append(i)
    return i_list


def skl_imputation(X, nominal_attributes):

    # Init
    n_rows, n_cols = X.shape
    i_list = []

    # Make imputers
    for c in range(n_cols):

        # Generate config
        i_config = dict(missing_values=np.nan)
        if c in nominal_attributes:
            # Nominal attribute
            out_kind = "nominal"
            i_config["strategy"] = "most_frequent"
        else:
            # Numeric attribute
            i_config["strategy"] = "mean"
            out_kind = "numeric"

        # Initialize imputer
        i = SimpleImputer(**i_config)
        i.fit(X[:, [c]])
        i.out_kind = out_kind
        i_list.append(i)

    return i_list



import itertools
import warnings
from inspect import signature
from timeit import default_timer

from sklearn.preprocessing import normalize

import dask
import numpy as np
import shap
from dask import delayed
from networkx import NetworkXUnfeasible, find_cycle, topological_sort
from sklearn.ensemble import (
    ExtraTreesClassifier,
    ExtraTreesRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.impute import SimpleImputer
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from ..algo import (
    evaluation,
    imputation,
    inference,
    inference_v3,
    new_inference,
    new_prediction,
    selection,
    vector_prediction,
)
from ..algo.induction import base_induction_algorithm, expand_induction_algorithm
from ..composition import CompositeModel, NewCompositeModel, o, x
from ..graph import build_diagram, compose_all, get_targ, model_to_graph
from ..utils import (
    DESC_ENCODING,
    MISS_ENCODING,
    TARG_ENCODING,
    DecoratedDecisionTreeClassifier,
    DecoratedDecisionTreeRegressor,
    DecoratedRandomForestClassifier,
    DecoratedRandomForestRegressor,
    code_to_query,
    get_i_o,
    query_to_code,
)
from ..visuals import save_diagram, show_diagram

try:
    from xgboost import XGBClassifier as XGBC
    from xgboost import XGBRegressor as XGBR
except:
    XGBC, XGBR = None, None

try:
    from lightgbm import LGBMClassifier as LGBMC
    from lightgbm import LGBMRegressor as LGBMR
except:
    LGBMC, LGBMR = None, None

try:
    from catboost import CatBoostClassifier as CBC
    from catboost import CatBoostRegressor as CBR
except:
    CBC, CBR = None, None

try:
    from wekalearn import RandomForestClassifier as WLC
    from wekalearn import RandomForestRegressor as WLR
except:
    WLC, WLR = None, None


class Mercs(object):
    delimiter = "_"

    selection_algorithms = dict(
        default=selection.base_selection_algorithm,
        base=selection.base_selection_algorithm,
        random=selection.random_selection_algorithm,
    )

    induction_algorithms = dict(
        base=base_induction_algorithm,
        default=base_induction_algorithm,
        expand=expand_induction_algorithm,
    )

    classifier_algorithms = dict(
        DT=DecisionTreeClassifier,
        DDT=DecoratedDecisionTreeClassifier,
        RF=RandomForestClassifier,
        DRF=DecoratedRandomForestClassifier,
        XGB=XGBC,
        xgb=XGBC,
        weka=WLC,
        LGBM=LGBMC,
        lgbm=LGBMC,
        CB=CBC,
        extra=ExtraTreesClassifier,
    )

    regressor_algorithms = dict(
        DT=DecisionTreeRegressor,
        DDT=DecoratedDecisionTreeRegressor,
        RF=RandomForestRegressor,
        DRF=DecoratedDecisionTreeRegressor,
        XGB=XGBR,
        xgb=XGBR,
        weka=WLR,
        LGBM=LGBMR,
        lgbm=LGBMR,
        CB=CBR,
        extra=ExtraTreesRegressor,
    )

    prediction_algorithms = dict(
        mi=vector_prediction.mi,
        mrai=vector_prediction.mrai,
        it=vector_prediction.it,
        rw=vector_prediction.rw,
    )

    inference_algorithms = dict(
        base=inference.base_inference_algorithm,
        dask=inference_v3.inference_algorithm,
        own=inference_v3.inference_algorithm,
    )

    imputer_algorithms = dict(
        nan=imputation.nan_imputation,
        NAN=imputation.nan_imputation,
        NaN=imputation.nan_imputation,
        null=imputation.nan_imputation,
        NULL=imputation.nan_imputation,
        skl=imputation.skl_imputation,
        base=imputation.skl_imputation,
        default=imputation.skl_imputation,
    )

    evaluation_algorithms = dict(
        base=evaluation.base_evaluation,
        default=evaluation.base_evaluation,
        dummy=evaluation.dummy_evaluation,
    )

    # Used in parse kwargs to identify parameters. If this identification goes wrong, you are sending settings
    # somewhere you do not want them to be. So, this is a tricky part, and moreover hardcoded. In other words:
    # this is risky terrain, and should probably be done differently in the future.
    configuration_prefixes = dict(
        imputation={"imputation", "imp"},
        induction={"induction", "ind"},
        selection={"selection", "sel"},
        prediction={"prediction", "pred", "prd"},
        inference={"inference", "infr", "inf"},
        classification={"classification", "classifier", "clf"},
        regression={"regression", "regressor", "rgr"},
        metadata={"metadata", "meta", "mtd"},
        evaluation={"evaluation", "evl"},
    )

    def __init__(
        self,
        selection_algorithm="base",
        induction_algorithm="base",
        classifier_algorithm="DT",
        regressor_algorithm="DT",
        prediction_algorithm="mi",
        inference_algorithm="own",
        imputer_algorithm="default",
        evaluation_algorithm="default",
        random_state=42,
        **kwargs
    ):
        self.params = dict(
            selection_algorithm=selection_algorithm,
            induction_algorithm=induction_algorithm,
            classifier_algorithm=classifier_algorithm,
            regressor_algorithm=regressor_algorithm,
            prediction_algorithm=prediction_algorithm,
            inference_algorithm=inference_algorithm,
            imputer_algorithm=imputer_algorithm,
            evaluation_algorithm=evaluation_algorithm,
            random_state=random_state,
        )
        self.params = {**self.params, **kwargs}

        self.random_state = random_state
        self.selection_algorithm = self.selection_algorithms[selection_algorithm]

        # N.b.: First try to look up the key. If the key is not found, we assume the algorithm itself was passed.
        self.classifier_algorithm = self.classifier_algorithms.get(
            classifier_algorithm, classifier_algorithm
        )
        self.regressor_algorithm = self.regressor_algorithms.get(
            regressor_algorithm, regressor_algorithm
        )

        self.prediction_algorithm = self.prediction_algorithms[prediction_algorithm]
        self.inference_algorithm = self.inference_algorithms[inference_algorithm]
        self.induction_algorithm = self.induction_algorithms[
            induction_algorithm
        ]  # For now, we only have one.
        self.imputer_algorithm = self.imputer_algorithms[imputer_algorithm]
        self.evaluation_algorithm = self.evaluation_algorithms[evaluation_algorithm]

        # Data-structures
        self.m_codes = np.array([])
        self.m_list = []
        self.c_list = []
        self.g_list = []
        self.i_list = []

        self.m_fimps = np.array([])
        self.m_score = np.array([])

        self.FI = np.array([])
        self.targ_ids = np.array([])

        # Query-related things
        self.q_code = None
        self.q_desc_ids = None
        self.q_targ_ids = None
        self.q_diagram = None
        self.q_compose = None
        self.q_methods = []

        # Configurations
        self.imp_cfg = self._default_config(self.imputer_algorithm)
        self.ind_cfg = self._default_config(self.induction_algorithm)
        self.sel_cfg = self._default_config(self.selection_algorithm)
        self.clf_cfg = self._default_config(self.classifier_algorithm)
        self.rgr_cfg = self._default_config(self.regressor_algorithm)
        self.prd_cfg = self._default_config(self.prediction_algorithm)
        self.inf_cfg = self._default_config(self.inference_algorithm)
        self.evl_cfg = self._default_config(self.evaluation_algorithm)

        self.configuration = dict(
            imputation=self.imp_cfg,
            induction=self.ind_cfg,
            selection=self.sel_cfg,
            classification=self.clf_cfg,
            regression=self.rgr_cfg,
            prediction=self.prd_cfg,
            inference=self.inf_cfg,
        )  # Collect all configs in one

        self._update_config(random_state=random_state, **kwargs)

        self.metadata = dict()
        self.model_data = dict()

        self._extra_checks_on_config()

        return

    def fit(self, X, y=None, m_codes=None, **kwargs):
        assert isinstance(X, np.ndarray)

        if y is not None:
            assert isinstance(y, np.ndarray)
            X = np.c_[X, y]

        tic = default_timer()

        self.metadata = self._default_metadata(X)
        self._update_metadata(**kwargs)

        self.i_list = self.imputer_algorithm(X, self.metadata.get("nominal_attributes"))

        # N.b.: `random state` parameter is in `self.sel_cfg`
        if m_codes is None:
            self.m_codes = self.selection_algorithm(self.metadata, **self.sel_cfg)
        else:
            self.m_codes = m_codes

        self.m_list = self.induction_algorithm(
            X,
            self.m_codes,
            self.metadata,
            self.classifier_algorithm,
            self.regressor_algorithm,
            self.clf_cfg,
            self.rgr_cfg,
            **self.ind_cfg
        )

        self._filter_m_list_m_codes()

        self._consistent_datastructures()

        if self.imputer_algorithm == self.imputer_algorithms.get("nan"):
            # If you do no have imputers, you cannot use them as a baseline evaluation
            self.evl_cfg["consider_imputations"] = False

        self.m_score = self.evaluation_algorithm(
            X, self.m_codes, self.m_list, self.i_list, **self.evl_cfg
        )

        toc = default_timer()
        self.model_data["ind_time"] = toc - tic
        self.metadata["n_component_models"] = len(self.m_codes)
        return

    def predict(
        self,
        X,
        q_code=None,
        inference_algorithm=None,
        prediction_algorithm=None,
        **kwargs
    ):
        # Update configuration if necessary
        if q_code is None:
            q_code = self._default_q_code()

        if inference_algorithm is not None:
            self._reconfig_inference(inference_algorithm=inference_algorithm)

        if prediction_algorithm is not None:
            self._reconfig_prediction(
                prediction_algorithm=prediction_algorithm, **kwargs
            )

        # Adjust data
        self.q_code = q_code
        self.q_desc_ids, self.q_targ_ids, _ = code_to_query(
            self.q_code, return_list=True
        )

        # Make query-diagram
        tic_prediction = default_timer()
        self.m_sel = self.prediction_algorithm(
            self.m_codes, self.m_fimps, self.m_score, q_code=self.q_code, **self.prd_cfg
        )
        toc_prediction = default_timer()

        tic_diagram = default_timer()
        self.q_diagram = self._build_q_diagram(self.m_list, self.m_sel)
        toc_diagram = default_timer()

        tic_infalgo = default_timer()
        if isinstance(self.q_diagram, tuple):
            self.q_diagrams = self.q_diagram

            # for d in self.q_diagrams:
            #    print(d.nodes)
            #   self.c_list.append(self._build_q_model(X, d))

            self.c_list = [self._build_q_model(X, d) for d in self.q_diagrams]
            self.c_sel = list(range(len(self.c_list)))
            self.c_diagram = self._build_q_diagram(
                self.c_list, self.c_sel, composition=True
            )

            self.q_model = self._build_q_model(X, self.c_diagram)
        else:
            self.q_model = self._build_q_model(X, self.q_diagram)

        toc_infalgo = default_timer()

        tic_dask = default_timer()
        X = X[:, self.q_model.desc_ids]
        result = self.q_model.predict(X)
        toc_dask = default_timer()

        self.model_data["prd_time"] = toc_prediction - tic_prediction
        self.model_data["dia_time"] = toc_diagram - tic_diagram
        self.model_data["infalgo_time"] = toc_infalgo - tic_infalgo
        self.model_data["dsk_time"] = toc_dask - tic_dask
        self.model_data["inf_time"] = toc_dask - tic_prediction

        return result

    def get_params(self, deep=False):
        return self.params

    # Diagrams
    def _build_q_diagram(self, m_list, m_sel, composition=False):
        if isinstance(m_sel, tuple):
            diagrams = [
                build_diagram(
                    m_list,
                    m_sel_instance,
                    self.q_code,
                    prune=True,
                    composition=composition,
                )
                for m_sel_instance in m_sel
            ]
            return tuple(diagrams)
        else:
            return build_diagram(
                m_list, m_sel, self.q_code, prune=True, composition=composition
            )

    def show_q_diagram(self, kind="svg", fi=False, ortho=False, index=None, **kwargs):

        if isinstance(self.q_diagram, tuple) and index is None:
            return show_diagram(self.c_diagram, kind=kind, fi=fi, ortho=ortho, **kwargs)
        elif isinstance(self.q_diagram, tuple):
            return show_diagram(
                self.q_diagram[index], kind=kind, fi=fi, ortho=ortho, **kwargs
            )
        else:
            return show_diagram(self.q_diagram, kind=kind, fi=fi, ortho=ortho, **kwargs)

    def save_diagram(self, fname=None, kind="svg", fi=False, ortho=False):
        return save_diagram(self.q_diagram, fname, kind=kind, fi=fi, ortho=ortho)

    # Inference
    def _build_q_model(self, X, diagram):
        try:
            self.inference_algorithm(
                diagram,
                self.m_list,
                self.i_list,
                self.c_list,
                X,
                self.metadata.get("nominal_attributes"),
            )
        except NetworkXUnfeasible:
            cycle = find_cycle(self.q_diagram, orientation="original")
            msg = """
            Topological sort failed, investigate diagram to debug.
            
            I will never be able to squeeze a prediction out of a diagram with a loop.
            
            Cycle was:  {}
            """.format(
                cycle
            )
            raise RecursionError(msg)

        n_component_models = self.metadata["n_component_models"]
        q_model = NewCompositeModel(
            diagram,
            nominal_attributes=self.metadata["nominal_attributes"],
            n_component_models=n_component_models,
        )
        return q_model

    def _merge_q_models(self, q_models):
        q_diagram = build_diagram(self.c_list, self.c_sel, self.q_code, prune=True)
        return q_diagram

    def merge_models(self, q_models):

        types = self._get_types(self.metadata)

        walks = [
            model_to_graph(m, types, idx=idx, composition=True)
            for idx, m in enumerate(q_models)
        ]
        q_diagram = compose_all(walks)
        filtered_nodes = self.filter_nodes(q_diagram)

        try:
            self.inference_algorithm(q_diagram, sorted_nodes=filtered_nodes)
        except NetworkXUnfeasible:
            cycle = find_cycle(q_diagram, orientation="original")
            msg = """
            Topological sort failed, investigate diagram to debug.
            
            I will never be able to squeeze a prediction out of a diagram with a loop.
            
            Cycle was:  {}
            """.format(
                cycle
            )
            raise RecursionError(msg)

        q_model = CompositeModel(q_diagram)
        return q_diagram, q_model

    def _get_q_model(self, q_diagram, X):

        self._add_imputer_function(q_diagram)

        try:
            self.inference_algorithm(q_diagram, X=X)
        except NetworkXUnfeasible:
            cycle = find_cycle(q_diagram, orientation="original")
            msg = """
            Topological sort failed, investigate diagram to debug.
            
            I will never be able to squeeze a prediction out of a diagram with a loop.
            
            Cycle was:  {}
            """.format(
                cycle
            )
            raise RecursionError(msg)

        q_model = CompositeModel(q_diagram)
        return q_model

    # Filter
    def _filter_m_list_m_codes(self):
        """Filtering out the failed models.

        This happens when TODO: EXPLAIN
        """

        fail_m_idxs = [i for i, m in enumerate(self.m_list) if m is None]
        self.m_codes = np.delete(self.m_codes, fail_m_idxs, axis=0)
        self.m_list = [m for m in self.m_list if m is not None]

        return

    # Graphs
    def _consistent_datastructures(self, binary_scores=False):
        self._update_m_codes()
        self._update_m_fimps()
        return

    def _expand_m_list(self):
        self.m_list = list(itertools.chain.from_iterable(self.m_list))
        return

    def _add_model(self, model, binary_scores=False):
        self.m_list.append(model)
        self._consistent_datastructures(binary_scores=binary_scores)
        return

    def _update_m_codes(self):
        self.m_codes = np.array(
            [
                query_to_code(
                    list(model.desc_ids),
                    list(model.targ_ids),
                    attributes=self.metadata["attributes"],
                )
                for model in self.m_list
            ]
        )
        return

    def _update_m_fimps(self):

        init = np.zeros(self.m_codes.shape)

        for m_idx, mod in enumerate(self.m_list):
            init[m_idx, list(mod.desc_ids)] = mod.feature_importances_

        self.m_fimps = init

        return

    def _update_m_score(self, binary_scores=False):
        if binary_scores:
            self.m_score = (self.m_codes == TARG_ENCODING).astype(float)
        return

    # Imputer
    def _add_imputer_function(self, g):

        for n in g.nodes:
            if g.nodes[n]["kind"] == "imputation":
                idx = g.nodes[n]["idx"]

                f_1 = self._dummy_array  # Artificial input
                f_2 = self.i_list[idx].transform  # Actual imputation
                f_3 = np.ravel  # Return a vector, not array

                g.nodes[n]["function"] = o(f_3, o(f_2, f_1))

        return

    # Add ids
    @staticmethod
    def _add_ids(g, desc_ids, targ_ids):
        g.graph["desc_ids"] = set(desc_ids)
        g.graph["targ_ids"] = set(targ_ids)
        return g

    # Metadata
    def _default_metadata(self, X):
        if X.ndim != 2:
            X = X.reshape(-1, 1)

        n_rows, n_cols = X.shape

        types = [X[0, 0].dtype for _ in range(n_cols)]
        nominal_attributes = set(
            [att for att, typ in enumerate(types) if self._is_nominal(typ)]
        )
        numeric_attributes = set(
            [att for att, typ in enumerate(types) if self._is_numeric(typ)]
        )

        metadata = dict(
            attributes=set(range(n_cols)),
            n_attributes=n_cols,
            types=types,
            nominal_attributes=nominal_attributes,
            numeric_attributes=numeric_attributes,
        )
        return metadata

    def _update_metadata(self, **kwargs):

        self._update_dictionary(self.metadata, kind="metadata", **kwargs)

        # Assure every attribute is `typed`: If not every attribute is here, set to numeric type (default)
        numeric = self.metadata["numeric_attributes"]
        nominal = self.metadata["nominal_attributes"]
        att_ids = self.metadata["attributes"]

        # All attributes should be accounted for and none should be double.
        if (len(nominal) + len(numeric) - len(att_ids)) != 0:
            numeric = att_ids - nominal
            self._update_dictionary(
                self.metadata, kind="metadata", numeric_attributes=numeric
            )

        return

    # Configuration
    def _reconfig_prediction(self, prediction_algorithm="mi", **kwargs):
        self.prediction_algorithm = self.prediction_algorithms[prediction_algorithm]
        self.prd_cfg = self._default_config(self.prediction_algorithm)

        self.configuration["prediction"] = self.prd_cfg
        self._update_config(**kwargs)

        return

    def _reconfig_inference(self, inference_algorithm="own", **kwargs):
        self.inference_algorithm = self.inference_algorithms[inference_algorithm]

        self.inf_cfg = self._default_config(self.inference_algorithm)

        self.configuration["inference"] = self.inf_cfg
        self._update_config(**kwargs)
        return

    @staticmethod
    def _default_config(method):
        config = {}
        sgn = signature(method)

        for key, parameter in sgn.parameters.items():
            if parameter.default is not parameter.empty:
                config[key] = parameter.default
        return config

    def _update_config(self, **kwargs):

        for kind in self.configuration:
            self._update_dictionary(self.configuration[kind], kind=kind, **kwargs)

        return

    def _extra_checks_on_config(self):

        self._check_xgb_single_target()
        return

    def _check_xgb_single_target(self):

        nb_targets = self.configuration["selection"]["nb_targets"]

        if nb_targets == 1:
            return None
        else:
            if (
                self.classifier_algorithm is self.classifier_algorithms["XGB"]
                or self.regressor_algorithm is self.regressor_algorithms["XGB"]
            ):
                xgb = True
            else:
                xgb = False

            if xgb:
                msg = """
                XGBoost cannot deal with multi-target outputs.

                Hence, the `nb_targets` parameter is automatically adapted to 1,
                so only single-target trees will be learned. 

                Please take this into account.
                """
                warnings.warn(msg)
                self.configuration["selection"]["nb_targets"] = 1

            return

    def _parse_kwargs(self, kind="selection", **kwargs):

        prefixes = [e + self.delimiter for e in self.configuration_prefixes[kind]]

        parameter_map = {
            x.split(prefix)[1]: x
            for x in kwargs
            for prefix in prefixes
            if x.startswith(prefix)
        }

        return parameter_map

    def _update_dictionary(self, dictionary, kind=None, **kwargs):
        # Immediate matches
        overlap = set(dictionary).intersection(set(kwargs))

        for k in overlap:
            dictionary[k] = kwargs[k]

        if kind is not None:
            # Parsed matches
            parameter_map = self._parse_kwargs(kind=kind, **kwargs)
            overlap = set(dictionary).intersection(set(parameter_map))

            for k in overlap:
                dictionary[k] = kwargs[parameter_map[k]]
        return

    # Helpers
    def _filter_X(self, X):
        # Filter relevant input attributes
        if X.shape[1] != len(self.q_compose.desc_ids):
            indices = self._overlapping_indices(
                self.q_desc_ids, self.q_compose.desc_ids
            )
        return X[:, indices]

    @staticmethod
    def _dummy_array(X):
        """
        Return an array of np.nan, with the same number of rows as the input array.

        Parameters
        ----------
        X:      np.ndarray(), n_rows, n_cols = X.shape,
                We use the shape of X to deduce shape of our output.

        Returns
        -------
        a:      np.ndarray(), shape= (n_rows, 1)
                n_rows is the same as the number of rows as X.

        """
        n_rows, _ = X.shape

        a = np.empty((n_rows, 1))
        a.fill(np.nan)

        return a

    def _default_q_code(self):

        q_code = np.zeros(self.metadata["n_attributes"])
        q_code[-1] = TARG_ENCODING

        return q_code

    @staticmethod
    def _is_nominal(t):
        condition_01 = t == np.dtype(int)
        return condition_01

    @staticmethod
    def _is_numeric(t):
        condition_01 = t == np.dtype(float)
        return condition_01

    @staticmethod
    def _get_types(metadata):
        nominal = {i: "nominal" for i in metadata["nominal_attributes"]}
        numeric = {i: "numeric" for i in metadata["numeric_attributes"]}
        return {**nominal, **numeric}

    @staticmethod
    def _overlapping_indices(a, b):
        """
        Given an array a and b, return the indices (in a) of elements that occur in both a and b.

        Parameters
        ----------
        a
        b

        Returns
        -------

        Examples
        --------
        a = [4,5,6]
        b = [4,6,7]

        overlapping_indices(a, b) = [0,2]

        """
        return np.nonzero(np.in1d(a, b))[0]

    @staticmethod
    def filter_nodes(g):
        # This is not as safe as it should be

        sorted_nodes = list(topological_sort(g))
        filtered_nodes = []
        for n in reversed(sorted_nodes):
            if g.nodes[n]["kind"] == "model":
                break
            filtered_nodes.append(n)
        filtered_nodes = list(reversed(filtered_nodes))
        return filtered_nodes

    # SYNTH
    def autocomplete(self, X, **kwargs):
        return

    # Legacy (delete when I am sure they can go)
    def predict_old(
        self, X, q_code=None, prediction_algorithm=None, beta=False, **kwargs
    ):
        # Update configuration if necessary
        if q_code is None:
            q_code = self._default_q_code()

        if prediction_algorithm is not None:
            reuse = False
            self._reconfig_prediction(
                prediction_algorithm=prediction_algorithm, **kwargs
            )

        # Adjust data
        tic_prediction = default_timer()
        self.q_code = q_code
        self.q_desc_ids, self.q_targ_ids, _ = code_to_query(
            self.q_code, return_list=True
        )

        # Make query-diagram
        self.q_diagram = self.prediction_algorithm(
            self.g_list, q_code, self.fi, self.t_codes, **self.prd_cfg
        )

        toc_prediction = default_timer()

        tic_dask = default_timer()

        toc_dask = default_timer()

        tic_compute = default_timer()
        res = self.q_model.predict.compute()
        toc_compute = default_timer()

        # Diagnostics
        self.model_data["prd_time"] = toc_prediction - tic_prediction
        self.model_data["dsk_time"] = toc_dask - tic_dask
        self.model_data["cmp_time"] = toc_compute - tic_compute
        self.model_data["inf_time"] = toc_compute - tic_prediction
        self.model_data["ratios"] = (
            self.model_data["prd_time"] / self.model_data["inf_time"],
            self.model_data["dsk_time"] / self.model_data["inf_time"],
            self.model_data["cmp_time"] / self.model_data["inf_time"],
        )
        return res

    def _update_g_list(self):
        types = self._get_types(self.metadata)
        self.g_list = [
            model_to_graph(m, types=types, idx=idx) for idx, m in enumerate(self.m_list)
        ]
        return

    def _update_t_codes(self):
        self.t_codes = (self.m_codes == TARG_ENCODING).astype(int)
        return

    # AVATAR-TOOLS
    def avatar(
        self,
        explainer_data,
        background_data=None,
        check_additivity=True,
        keep_abs_shaps=False,
        **explainer_kwargs
    ):

        self._init_avatar()

        for m_idx in range(len(self.m_list)):
            # Extract tree and m_code
            tree = self.m_list[m_idx].model
            m_code = self.m_codes[m_idx]

            # Filter data
            attribute_filter = m_code == DESC_ENCODING
            X = explainer_data[:, attribute_filter]

            if background_data is not None:
                B = background_data[:, attribute_filter]
            else:
                B = background_data

            # Shap Calculation
            explainer = shap.TreeExplainer(
                tree, data=B, **explainer_kwargs
            )
            raw_shaps = explainer.shap_values(X, check_additivity=check_additivity)

            # Process Shap values
            tsr_shaps = np.array(raw_shaps)  # tensor
            abs_shaps = np.abs(tsr_shaps)  # absolute

            if len(abs_shaps.shape) == 3:
                # In case of nominal target, sum shap values across target classes
                abs_shaps = np.sum(abs_shaps, axis=0)

            avg_shaps = np.mean(
                abs_shaps, axis=0
            )  # Avg over instances (of explainer data!)

            nrm_shaps = np.squeeze(
                normalize(avg_shaps.reshape(1, -1), norm="l1")
            )  # Normalize (between 0 and 1)

            if keep_abs_shaps:
                self.abs_shaps.append(abs_shaps)
            self.nrm_shaps.append(nrm_shaps)

        self._format_abs_shaps()
        self._format_nrm_shaps()

        return

    def _init_avatar(self):
        """Initialize avatar-datastructures that are used there.
        """
        self.abs_shaps = []
        self.nrm_shaps = []
        return

    def _format_nrm_shaps(self):
        if isinstance(self.nrm_shaps, list) and len(self.nrm_shaps) > 0:
            init = np.zeros(self.m_codes.shape)

            for m_idx, (mod, nrm_shap) in enumerate(zip(self.m_list, self.nrm_shaps)):
                init[m_idx, list(mod.desc_ids)] = nrm_shap

            self.nrm_shaps = init
        else:
            return

    def _format_abs_shaps(self):
        if isinstance(self.abs_shaps, list) and len(self.abs_shaps) > 0:
            n_models, n_attributes = self.m_codes.shape
            n_instances = self.abs_shaps[0].shape[0]
            init = np.zeros((n_models, n_instances, n_attributes))

            for m_idx, (mod, abs_shap) in enumerate(zip(self.m_list, self.abs_shaps)):
                init_abs = np.zeros((n_instances, n_attributes))
                init_abs[:, list(mod.desc_ids)] = abs_shap
                init[m_idx, :, :] = init_abs

            self.abs_shaps = init
        else:
            return




from .network import (
    model_to_graph,
    add_merge_nodes,
    add_imputation_nodes,
    compose,
    get_ids,
    get_nodes,
    compose_all,
    get_targ,
    get_desc
)

from .graphviz import fix_layout, to_dot

from  .q_diagram import build_diagram



import warnings
from functools import reduce

import networkx as nx
import numpy as np
from networkx import NetworkXNoCycle
from networkx.algorithms import find_cycle

from ..graph import add_imputation_nodes, add_merge_nodes, compose_all, get_ids
from ..utils import (
    change_role,
    debug_print,
    get_att,
    ENCODING,
    DESC_ENCODING,
    TARG_ENCODING,
    MISS_ENCODING,
)

VERBOSITY = 0


def mi(g_list, q_code, fi, t_codes, random_state=997):
    return mrai(
        g_list,
        q_code,
        fi,
        t_codes,
        init_threshold=-0.1,
        stepsize=0.1,
        greedy=True,
        stochastic=False,
        any_target=False,
        imputation_nodes=True,
        merge_nodes=True,
        return_avl_g=False,
        random_state=random_state,
    )


def mi_old(g_list, q_code, fi, t_codes, random_state=997):

    # Init
    mod_ids = [g.graph["id"] for g in g_list]

    # Calculate criteria
    avl_att = _att_indicator(q_code, kind="desc")
    avl_mod = _mod_indicator(mod_ids, q_code, t_codes, any_target=True)
    criteria = avl_mod

    # Pick
    g_sel = _avl_pick(g_list, criteria)

    # Build new graph
    q_diagram = _build_diagram(
        g_sel, avl_att, imputation_nodes=True, merge_nodes=True, test=True
    )

    return q_diagram


def mrai(
    g_list,
    q_code,
    fi,
    t_codes,
    init_threshold=1.0,
    stepsize=0.1,
    greedy=True,
    stochastic=False,
    any_target=False,
    imputation_nodes=True,
    merge_nodes=True,
    return_avl_g=False,
    random_state=997,
):

    # Init
    thresholds = _init_thresholds(init_threshold, stepsize)

    mod_ids = [g.graph["id"] for g in g_list]

    # Calculate criteria
    avl_att = _att_indicator(q_code, kind="desc")
    avl_mod = _mod_indicator(mod_ids, q_code, t_codes, any_target=any_target)
    criteria = _mod_criteria(mod_ids, avl_att, fi)

    criteria = (
        criteria * avl_mod
    )  # All the criteria of unavailable models are set to zero

    criteria[np.where(avl_mod<=0)[0], :] = -1  # Unavailable models are set to -1, available ones keep their score.

    # Pick
    for c_idx in range(criteria.shape[1]):
        criterion = criteria[:, c_idx]

        if greedy:
            g_sel = _greedy_pick(g_list, criterion, thresholds)
        elif stochastic:
            g_sel = _stochastic_pick(g_list, criterion, n=1, random_state=random_state)
        else:
            msg = """
            You either need to pick in a greedy fashion (picking the most appropriate models first) or
            in a stochastic fashion (more likely to pick the most appropriate models). Both of those
            options evaluated to False so I have no idea what you are trying to do.
            """
            raise NotImplementedError(msg)

    # Build new graph
    q_diagram = _build_diagram(
        g_sel,
        avl_att,
        imputation_nodes=imputation_nodes,
        merge_nodes=merge_nodes,
        test=True,
    )

    if return_avl_g:
        sel_ids = [g.graph["id"] for g in g_sel]
        avl_g = [g for g in g_list if g.graph["id"] not in sel_ids]
        return q_diagram, avl_g
    else:
        return q_diagram


def it(
    g_list,
    q_code,
    fi,
    t_codes,
    max_steps=4,
    init_threshold=1.0,
    stepsize=0.1,
    random_state=997,
):
    # Init
    stochastic = False
    greedy = True
    any_target = True

    g_sel = []
    avl_att = _att_indicator(q_code, kind="desc")
    tgt_att = _att_indicator(q_code, kind="targ")

    step_g_list = g_list
    step_q_code = change_role(q_code, MISS_ENCODING, TARG_ENCODING)

    for step in range(max_steps):

        last = step + 1 == max_steps  # Check if this is our last chance
        if last:
            any_target = False  # We finish the job
            q_targ_todo = _targ_todo(step_q_code, tgt_att)
            step_q_code = change_role(step_q_code, TARG_ENCODING, MISS_ENCODING)
            step_q_code[q_targ_todo] = TARG_ENCODING

        step_q_diagram, step_g_list = mrai(
            step_g_list,
            step_q_code,
            fi,
            t_codes,
            init_threshold=init_threshold,
            stepsize=stepsize,
            greedy=greedy,
            stochastic=stochastic,
            any_target=any_target,
            imputation_nodes=True,
            merge_nodes=False,
            return_avl_g=True,
            random_state=random_state,
        )

        # Remember graph of this step
        g_sel.append(step_q_diagram)

        # Update query
        step_targ = list(step_q_diagram.graph["targ_ids"])
        step_q_code[step_targ] = DESC_ENCODING

        if _stopping_criterion_it(step_q_code, tgt_att):
            break

    # Build diagram
    q_diagram = _build_diagram(
        g_sel,
        avl_att,
        tgt_att,
        imputation_nodes=False,
        merge_nodes=True,
        prune=True,
        test=False,
    )
    return q_diagram


def rw(
    g_list,
    q_code,
    fi,
    t_codes,
    max_steps=4,
    nb_walks=1,
    init_threshold=1.0,
    stepsize=0.1,
    random_state=997,
):

    q_diagrams = [
        walk(
            g_list,
            q_code,
            fi,
            t_codes,
            max_steps=max_steps,
            init_threshold=init_threshold,
            stepsize=stepsize,
            random_state=random_state + i,
        )
        for i in range(nb_walks)
    ]

    return q_diagrams


def walk(
    g_list,
    q_code,
    fi,
    t_codes,
    max_steps=4,
    init_threshold=1.0,
    stepsize=0.1,
    random_state=997,
):

    # Init
    avl_att = _att_indicator(q_code, kind="desc")
    tgt_att = _att_indicator(q_code, kind="targ")
    stochastic = True
    greedy = False
    any_target = True

    g_sel = []
    g_tgt = set([])
    q_desc = set(get_att(q_code, kind="desc"))
    q_miss = set(get_att(q_code, kind="miss"))

    step_g_list = g_list
    step_q_code = q_code.copy()

    # Generate chain
    for step in reversed(range(max_steps)):
        step_q_diagram, step_g_list = mrai(
            step_g_list,
            step_q_code,
            fi,
            t_codes,
            init_threshold=init_threshold,
            stepsize=stepsize,
            greedy=greedy,
            stochastic=stochastic,
            any_target=any_target,
            imputation_nodes=False,
            merge_nodes=True,
            return_avl_g=True,
            random_state=random_state,
        )

        # Update query
        step_targ = step_q_diagram.graph["targ_ids"]
        step_desc = step_q_diagram.graph["desc_ids"]

        # Remember graph of this step (Append in front!)
        g_sel.insert(0, step_q_diagram)
        g_tgt = g_tgt.union(step_targ)

        # Extract info
        step_q_code[:] = MISS_ENCODING
        step_q_code[list(q_desc)] = DESC_ENCODING
        step_q_code[list(q_miss.intersection(step_desc))] = TARG_ENCODING
        step_q_code[
            list(g_tgt)
        ] = (
            MISS_ENCODING
        )  # Consider future targets forbidden (this might not be necessary)

        if _stopping_criterion_rw(step_q_code, step):
            break

    # Build diagram
    avl_desc = set(q_desc)
    for g in g_sel:
        add_imputation_nodes(g, avl_desc)
        avl_desc = avl_desc.union(g.graph["targ_ids"])

    q_diagram = _build_diagram(
        g_sel,
        avl_att,
        tgt_att,
        imputation_nodes=False,
        merge_nodes=True,
        prune=True,
        test=False,
    )
    return q_diagram


def rev(
    g_list,
    q_code,
    fi,
    t_codes,
    max_steps=4,
    init_threshold=1.0,
    stepsize=0.1,
    random_state=997,
):

    # Init
    avl_att = _att_indicator(q_code, kind="desc")
    tgt_att = _att_indicator(q_code, kind="targ")
    stochastic = True
    greedy = False
    any_target = True

    g_sel = []
    g_tgt = set([])
    q_desc = set(get_att(q_code, kind="desc"))
    q_miss = set(get_att(q_code, kind="miss"))

    step_g_list = g_list
    step_q_code = q_code

    # Generate chain
    for step in reversed(range(max_steps)):
        step_q_diagram, step_g_list = mrai(
            step_g_list,
            step_q_code,
            fi,
            t_codes,
            init_threshold=init_threshold,
            stepsize=stepsize,
            greedy=greedy,
            stochastic=stochastic,
            any_target=any_target,
            imputation_nodes=False,
            merge_nodes=True,
            return_avl_g=True,
            random_state=random_state,
        )

        # Update query
        step_targ = step_q_diagram.graph["targ_ids"]
        step_desc = step_q_diagram.graph["desc_ids"]

        # Remember graph of this step (Append in front!)
        g_sel.insert(0, step_q_diagram)
        g_tgt = g_tgt.union(step_targ)

        # Extract info
        step_q_code[list(q_miss.intersection(step_desc))] = TARG_ENCODING
        step_q_code[
            list(g_tgt)
        ] = (
            MISS_ENCODING
        )  # Consider future targets forbidden (this might not be necessary)

        if _stopping_criterion_rw(step_q_code, step):
            break

    # Build diagram
    avl_desc = set(q_desc)
    for g in g_sel:
        print(g.nodes())
        print(g.graph["desc_ids"], g.graph["targ_ids"])
        add_imputation_nodes(g, avl_desc)
        avl_desc = avl_desc.union(g.graph["targ_ids"])

    q_diagram = _build_diagram(
        g_sel,
        avl_att,
        tgt_att,
        imputation_nodes=False,
        merge_nodes=True,
        prune=True,
        test=False,
    )
    return q_diagram


# Stopping criteria
def _stopping_criterion_it(step_q_code, tgt_att):
    q_targ_todo = _targ_todo(step_q_code, tgt_att)
    return not np.any(q_targ_todo)


def _stopping_criterion_rw(step_q_code, step):
    reason_01 = len(get_att(step_q_code, kind="targ")) == 0
    reason_02 = step == 0
    return reason_01 or reason_02


def _stopping_criterion_greedy_pick(list_of_graphs):
    return len(list_of_graphs) > 0


def _targ_todo(step_q_code, tgt_att):
    """
    Everything you want to know and do not know yet.
    """
    return (tgt_att & (step_q_code != DESC_ENCODING)).astype(bool)


# Criteria-calculations
def _mod_criteria(mod_ids, avl_att, fi):
    return np.dot(fi[mod_ids, :], avl_att).reshape(-1, 1)


def _mod_indicator(mod_ids, q_code, t_codes, any_target=True):
    """Indicator vector of available models.

    Available models are models which are eligible for selection by 
    the algorithm. Currently, this just means that the model has to be 
    relevant, it has to predict at least one target that is required by the query.

    Hence, changes in behaviour (e.g., predicting a missing attribute is also OK) have to
    be realized at the level of the query, not here.

    Parameters
    ----------
    mod_ids: 
    q_code: 
    t_codes: 

    Returns
    -------

    """

    avl_tgt = np.eye(q_code.shape[0], dtype=int)[q_code == TARG_ENCODING].T
    avl_mod = np.dot(t_codes[mod_ids, :], avl_tgt)

    if any_target:
        # If any target works, I can just sum them up.
        # Predicting multiple interesting targets will help to some extent! (sum)
        avl_mod = np.sum(avl_mod, axis=1, keepdims=True)
        
        ones_idx = np.where(avl_mod > 0.)
        avl_mod[:] = 0
        avl_mod[ones_idx] = 1

        #avl_mod = np.clip(np.max(avl_mod, axis=1, keepdims=True), a_min=0, a_max=1)

    return avl_mod


def _att_indicator(q_code, kind="desc"):
    return (q_code == ENCODING[kind]).astype(int)


# Pick models
def _stochastic_pick(g_list, criteria, n=1, random_state=997):
    """
    Interpret an array of appropriateness scores as a distribution
    corresponding to the probability of a certain model being selected.


    Parameters
    ----------
    criteria:   list
                Array that quantifies how likely a pick should be.
    n:          int
                Number of picks

    Returns
    -------
    picks:      np.ndarray
                List of indices that were picked
    """

    np.random.seed(random_state)
    criteria += abs(np.min([0, np.min(criteria)]))  # Shift in case of negative values
    norm = np.linalg.norm(criteria, 1)

    if norm > 0:
        criteria = criteria / norm
    else:
        msg = """
        Not a single appropriate model was found, therefore
        making an arbitrary choice.
        """
        warnings.warn(msg)
        # If you cannot be right, be arbitrary
        criteria = [1 / len(criteria) for i in criteria]
        

    draw = np.random.multinomial(1, criteria, size=n)
    picks = np.where(draw == 1)[1]

    return [g_list[g_idx].copy() for g_idx in picks]


def _greedy_pick(g_list, criteria, thresholds):
    # I do not think I need to copy!
    for thr in thresholds:
        g_sel = [g_list[idx].copy() for idx, c in enumerate(criteria) if c >= thr]
        if _stopping_criterion_greedy_pick(g_sel):
            break
    return g_sel


def _avl_pick(g_list, criteria):
    g_sel = [g_list[idx].copy() for idx, c in enumerate(criteria) if c > 0.8]
    return g_sel


# Initializations
def _init_thresholds(init_threshold, stepsize):
    """Initialize thresholds array based on its two defining parameters.

    Parameters
    ----------
    init_threshold: 
    stepsize: 

    Returns
    -------

    """

    thresholds = np.arange(init_threshold, -1 - stepsize, -stepsize)
    thresholds = np.clip(thresholds, -1, 1)
    return thresholds


# Graph


def _build_diagram(
    g_sel,
    avl_att,
    tgt_att=None,
    imputation_nodes=True,
    merge_nodes=False,
    prune=False,
    test=False,
):
    if imputation_nodes:
        q_desc = np.where(avl_att)[0]
        for g in g_sel:
            add_imputation_nodes(g, q_desc)

    q_diagram = compose_all(g_sel)

    if merge_nodes:
        add_merge_nodes(q_diagram)

    if prune:
        assert tgt_att is not None, "If you prune, you need to provide tgt_att"
        q_targ = np.where(tgt_att)[0]
        _prune(q_diagram, q_targ)

    if test:
        try:
            cycles = find_cycle(q_diagram)
            msg = """
            Found a cycle!
            Cycle was: {}
            """.format(
                cycles
            )
            raise ValueError(msg)
        except NetworkXNoCycle:
            pass

    q_diagram.graph["desc_ids"] = get_ids(q_diagram, kind="desc")
    q_diagram.graph["targ_ids"] = get_ids(q_diagram, kind="targ")

    return q_diagram


def _prune(g, tgt_nodes=None):

    msg = """
    tgt_nodes:          {}
    tgt_nodes[0]:       {}
    type(tgt_nodes[0]): {}
    """.format(
        tgt_nodes, tgt_nodes[0], type(tgt_nodes[0])
    )
    debug_print(msg, level=1, V=VERBOSITY)

    if tgt_nodes is None:
        tgt_nodes = [
            n
            for n, out_degree in g.out_degree()
            if out_degree == 0
            if g.nodes[n]["kind"] == "data"
        ]
        msg = """
        tgt_nodes: {}
        """.format(
            tgt_nodes
        )
        debug_print(msg, level=1, V=VERBOSITY)
    elif isinstance(tgt_nodes[0], (int, np.int64)):
        tgt_nodes = [
            n
            for n in g.nodes
            if g.nodes[n]["kind"] == "data"
            if g.nodes[n]["idx"] in tgt_nodes
        ]
    else:
        assert isinstance(tgt_nodes[0], str)

    ancestors = [nx.ancestors(g, source=n) for n in tgt_nodes]
    retain_nodes = reduce(set.union, ancestors, set(tgt_nodes))

    nodes_to_remove = [n for n in g.nodes if n not in retain_nodes]
    for n in nodes_to_remove:
        g.remove_node(n)
    return



from functools import partial

import networkx as nx
import numpy as np
from dask import delayed

from ..composition import o
from ..utils.inference_tools import (
    _dummy_array,
    _map_classes,
    _pad_proba,
    _select_nominal,
    _select_numeric,
)


# Main algorithm
def inference_algorithm(g, m_list, i_list, data, nominal_ids):

    data_node = lambda k, n: k == "D"
    model_node = lambda k, n: k == "M"
    imputation_node = lambda k, n: k == "I"

    nodes = list(nx.topological_sort(g))
    nb_rows, _ = data.shape

    g_desc_ids = list(g.desc_ids)
    data = data[:, g_desc_ids]

    for n in nodes:
        if data_node(*n):
            in_degree = g.in_degree(n)
            if in_degree == 0:
                dask_input_data_node(g, n, g_desc_ids, data)
            elif in_degree == 1:
                dask_single_data_node(g, n, m_list)
            elif in_degree > 1:
                if n[1] in nominal_ids:
                    dask_nominal_data_node(g, n, m_list)
                else:
                    dask_numeric_data_node(g, n, m_list)
        elif model_node(*n):
            dask_model_node(g, n, m_list)
        elif imputation_node(*n):
            dask_imputation_node(g, n, i_list, nb_rows)
        else:
            raise ValueError("Did not recognize node kind of {}".format(n))

    return


# Specific Nodes
def dask_input_data_node(g, node, g_desc_ids, data):

    f = _select_numeric(g_desc_ids.index(node[1]))

    g.node[node]["dask"] = delayed(f)(data)

    # New
    g.node[node]["inputs"] = data
    g.node[node]["compute"] = f

    return


def dask_imputation_node(g, node, i_list, nb_rows):

    # Build function
    f1 = _dummy_array
    f2 = i_list[node[1]].transform
    f3 = np.ravel
    f = o(f3, o(f2, f1))

    g.node[node]["dask"] = delayed(f)(nb_rows)

    # New
    g.node[node]["inputs"] = nb_rows
    g.node[node]["compute"] = f
    return


def dask_single_data_node(g, node, m_list):

    # Build function
    idx, parent_function = _dask_get_parents_of_numeric_data_node(g, m_list, node)[
        0
    ]  # Single input
    f = _select_numeric(idx)
    g.node[node]["dask"] = delayed(f)(parent_function)

    # New
    parents = _numeric_parents(g, m_list, node)

    def f(parents):
        collector = _numeric_inputs(parents)
        return collector.pop()

    g.node[node]["inputs"] = parents
    g.node[node]["compute"] = f

    return


def dask_numeric_data_node(g, node, m_list):

    idx_fnc = _dask_get_parents_of_numeric_data_node(g, m_list, node)

    parent_functions = [delayed(_select_numeric(idx))(fnc) for idx, fnc in idx_fnc]

    f1 = partial(np.mean, axis=0)

    g.node[node]["dask"] = delayed(f1)(parent_functions)

    # New
    parents = _numeric_parents(g, m_list, node)

    def f(parents):
        collector = _numeric_inputs(parents)
        return f1(collector)

    g.node[node]["inputs"] = parents
    g.node[node]["compute"] = f
    return


def dask_nominal_data_node(g, node, m_list):
    idx_cls_fnc = _dask_get_parents_of_nominal_data_node(g, m_list, node)
    classes = np.unique(np.hstack([c for _, c, _ in idx_cls_fnc]))

    # Reduce
    parent_functions = []

    for idx, c, fnc in idx_cls_fnc:
        f1 = delayed(_select_nominal(idx))(fnc)
        if len(c) < len(classes):
            f2 = delayed(_pad_proba(c, classes))(f1)
            parent_functions.append(f2)
        else:
            parent_functions.append(f1)

    f3 = delayed(partial(np.sum, axis=0))(parent_functions)
    g.node[node]["dask_proba"] = f3

    g.node[node]["classes"] = classes

    # Vote
    def vote(X):
        return classes.take(np.argmax(X, axis=1), axis=0)

    g.node[node]["dask"] = delayed(vote)(f3)

    # New
    """
    parents = (
        _select_nominal(idx)(fnc)
        if len(c) == len(classes)
        else _pad_proba(c, classes)(_select_nominal(idx)(fnc))
        for idx, c, fnc in _nominal_parents(g, m_list, node)
    )
    
    def F(parents):
        collector = np.sum(list(parents), axis=0)
        return collector

    def F2(parents):
        try:
            collector = np.sum(list(parents), axis=0)
            return vote(collector)
        except:
            collector = np.sum(list(parents), axis=0)
            return collector
    """

    parents = _nominal_parents(g, m_list, node)
    def F(parents):
        collector = _nominal_inputs(g, parents, classes)
        return np.sum(collector, axis=0)

    def F2(parents):
        return vote(F(parents))

    g.node[node]["inputs"] = parents
    g.node[node]["compute_proba"] = F
    g.node[node]["compute"] = F2

    return


def dask_model_node(g, node, m_list):
    # Collect input data
    parent_functions = _dask_get_parents_of_model_node(g, node)

    collector = delayed(np.stack, pure=True)(parent_functions, axis=1)

    # Convert function
    g.node[node]["dask"] = delayed(m_list[node[1]].predict)(collector)

    if hasattr(m_list[node[1]], "predict_proba"):
        g.node[node]["dask_proba"] = delayed(m_list[node[1]].predict_proba)(collector)

    # New
    parents = _model_parents(g, node)
    
    def f(parents):
        X = _model_inputs(g, parents)
        return m_list[node[1]].predict(X)

    g.node[node]["inputs"] = parents
    g.node[node]["compute"] = f

    if hasattr(m_list[node[1]], "predict_proba"):
        def f2(parents):
            X = _model_inputs(g, parents)
            return m_list[node[1]].predict_proba(X)

        g.node[node]["compute_proba"] = f2

    return


# Helpers - Function
def _nominal_inputs(g, parents, classes):
    collector = [
            _select_nominal(idx)(compute(g, n, proba=True))
            if len(c) == len(classes)
            else _pad_proba(c, classes)(_select_nominal(idx)(compute(g, n, proba=True)))
            for idx, c, n in parents
        ]
    return collector

def _numeric_inputs(g, parents):
    collector = [_select_numeric(idx)(compute(g, n)) for idx, n in parents]
    return collector

def _model_inputs(g, parents):
    collector = [compute(g, n) for n in parents]
    collector = np.stack(collector, axis=1)
    return collector

def compute(g, node, proba=False):

    result = "result"
    compute = "compute"

    if proba:
        result += "_proba"
        compute += "_proba"

    r = g.node[node].get(result, None)
    if r is None:
        i = g.node[node].get("inputs")
        f = g.node[node].get(compute)
        g.node[node][result] = f(i)
        return g.node[node][result]
    else:
        return r


def _numeric_parents(g, m_list, node):
    rel_idx = lambda p_idx, n_idx: m_list[p_idx].targ_ids.index(n_idx)

    # parents = (
    #    (rel_idx(p_idx, node[1]) if m == "M" else 0, compute(g, (m, p_idx)))
    #    for m, p_idx in g.predecessors(node)
    #

    parents = [
        (rel_idx(p_idx, node[1]) if m == "M" else 0, (m, p_idx))
        for m, p_idx in g.predecessors(node)
    ]

    return parents


def _nominal_parents(g, m_list, node):
    rel_idx = lambda p_idx, n_idx: m_list[p_idx].targ_ids.index(n_idx)
    classes = lambda p_idx, r_idx: m_list[p_idx].classes_[r_idx]

    # parents = (
    #    (
    #        rel_idx(p_idx, node[1]) if m == "M" else 0,
    #        classes(p_idx, rel_idx(p_idx, node[1]) if m == "M" else 0),
    #        compute(g, (m, p_idx), proba=True),
    #    )
    #    for m, p_idx in g.predecessors(node)
    # )

    parents = [
        (
            rel_idx(p_idx, node[1]) if m == "M" else 0,
            classes(p_idx, rel_idx(p_idx, node[1]) if m == "M" else 0),
            (m, p_idx),
        )
        for m, p_idx in g.predecessors(node)
    ]

    return parents


def _model_parents(g, node):
    idxs = {p_idx: (m, p_idx) for m, p_idx in g.predecessors(node)}

    # parents = (compute(g, n) for k, n in sorted(idxs.items()))

    parents = [n for k, n in sorted(idxs.items())]

    return parents


# Helpers - Dask
def _dask_get_parents_of_model_node(g, node):
    parent_functions = {a: g.nodes[(m, a)]["dask"] for m, a in g.predecessors(node)}
    parent_functions = [v for k, v in sorted(parent_functions.items())]
    return parent_functions


def _dask_get_parents_of_numeric_data_node(g, m_list, node):
    rel_idx = lambda p_idx, n_idx: m_list[p_idx].targ_ids.index(n_idx)

    idx_fnc = [
        (rel_idx(p_idx, node[1]) if m == "M" else 0, g.node[(m, p_idx)]["dask"])
        for m, p_idx in g.predecessors(node)
    ]

    return idx_fnc


def _dask_get_parents_of_nominal_data_node(g, m_list, node):
    rel_idx = lambda p_idx, n_idx: m_list[p_idx].targ_ids.index(n_idx)
    classes = lambda p_idx, r_idx: m_list[p_idx].classes_[r_idx]

    idx_fnc = (
        (
            rel_idx(p_idx, node[1]) if m == "M" else 0,
            p_idx,
            g.node[(m, p_idx)]["dask_proba"],
        )
        for m, p_idx in g.predecessors(node)
    )
    idx_cls_fnc = [(r_idx, classes(p_idx, r_idx), f) for r_idx, p_idx, f in idx_fnc]

    return idx_cls_fnc



import numpy as np
from dask import delayed
from sklearn.preprocessing import normalize

from ..algo.inference_v3 import compute
from .compose import o, x
from ..graph.network import get_ids, node_label, get_nodes

from ..utils import debug_print

VERBOSITY = 0


class NewCompositeModel(object):
    def __init__(self, diagram, desc_ids=None, targ_ids=None, nominal_attributes=None, n_component_models=0):

        # Assign desc and targ ids

        for k, idx in diagram.nodes():
            if k == 'M' and (idx >= n_component_models):
                diagram.nodes[(k, idx)]['shape'] = 'square'

        if desc_ids is not None:
            self.desc_ids = desc_ids
        else:
            self.desc_ids = diagram.desc_ids

        if targ_ids is not None:
            self.targ_ids = targ_ids
        else:
            self.targ_ids = diagram.targ_ids

        self.desc_ids = sorted(list(self.desc_ids))
        self.targ_ids = sorted(list(self.targ_ids))

        self.feature_importances_ = [1/len(self.desc_ids) for _ in self.desc_ids]

        self.predict = _get_predict(diagram, self.targ_ids)

        if nominal_attributes is not None:
            self.nominal_targ_ids = nominal_attributes.intersection(self.targ_ids)
            self.classes_ = _get_classes_(diagram, self.nominal_targ_ids)
            self.predict_proba = _get_predict_proba(diagram, self.nominal_targ_ids)

        self.score = 1

        return


    def get_confidences(self, X=None, redo=False, normalize_outputs=True):

        confidences = [np.array([[1.0]]) for t in self.targ_ids]
        nominal_prb = self.predict_proba(X, redo=redo)

        for targ_id, proba in zip(self.nominal_targ_ids, nominal_prb):
            targ_idx = self.targ_ids.index(targ_id)
            confidences[targ_idx] = proba

        if normalize_outputs:
            confidences = [np.max(normalize(c, norm='l1')) for c in confidences]
        return confidences


def _get_predict(diagram, targ_ids):
    def predict(X, redo=True):
        diagram.data = X

        if redo:
            clean_cache(diagram)

        collector = [compute(diagram, ("D", n)) for n in targ_ids]
        if len(targ_ids) == 1:
            return collector.pop()
        else:
            return np.stack(collector, axis=1)

    return predict


def _get_predict_proba(diagram, nominal_targ_ids):
    def predict_proba(X, redo=True):
        diagram.data = X

        if redo:
            clean_cache(diagram)

        collector = [compute(diagram, ("D", n), proba=True) for n in nominal_targ_ids]
        return collector

    return predict_proba


def _get_classes_(diagram, nominal_targ_ids):
    collector = [diagram.nodes[("D", n)]["classes"] for n in nominal_targ_ids]
    return collector


def clean_cache(diagram):
    for n in diagram.nodes:
        if diagram.nodes[n].get("result", None) is not None:
            diagram.nodes[n]["result"] = None
        if diagram.nodes[n].get("result_proba", None) is not None:
            diagram.nodes[n]["result_proba"] = None
    return



import warnings
from functools import partial, wraps
import itertools
import numpy as np
import pandas as pd

from sklearn.ensemble import (
    RandomForestClassifier,
    RandomForestRegressor,
    ExtraTreesClassifier,
    ExtraTreesRegressor,
)
from sklearn.metrics import f1_score, r2_score
from sklearn.model_selection import train_test_split
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from joblib import Parallel, delayed
from ..utils.decoration import decorate_tree

from sklearn.preprocessing import normalize
import shap


try:
    from xgboost import XGBClassifier as XGBC
    from xgboost import XGBRegressor as XGBR
except:
    XGBC, XGBR = None, None

try:
    from lightgbm import LGBMClassifier as LGBMC
    from lightgbm import LGBMRegressor as LGBMR
except:
    LGBMC, LGBMR = None, None

try:
    from catboost import CatBoostClassifier as CBC
    from catboost import CatBoostRegressor as CBR
except:
    CBC, CBR = None, None

try:
    from wekalearn import RandomForestClassifier as WLC
    from wekalearn import RandomForestRegressor as WLR
except:
    WLC, WLR = None, None

from ..composition.CanonicalModel import CanonicalModel
from ..utils import code_to_query, debug_print, get_att, get_i_o


def base_induction_algorithm(
    data,
    m_codes,
    metadata,
    classifier,
    regressor,
    classifier_kwargs,
    regressor_kwargs,
    random_state=997,
    calculation_method_feature_importances="default",
    min_nb_samples=10,
    n_jobs=1,
    verbose=0,
):
    """Basic induction algorithm. Models according to the m_codes it receives.
    
    Arguments:
        data {np.ndarray} -- Input data
        m_codes {np.ndarray} -- Model codes
        metadata {dict} -- Metadata of MERCS
        classifier {Supported ML learner} -- Supported ML learner
        regressor {Supported ML learner} -- Supported ML learner
        classifier_kwargs {dict} -- Kwargs for classifier
        regressor_kwargs {dict} -- Kwargs for regressor
    
    Keyword Arguments:
        random_state {int} -- Seed for random numbers (default: {997})
        n_jobs {int} -- Joblib can be used in training. (default: {1})
        verbose {int} -- Verbosity level for Joblib. (default: {0})
    
    Raises:
        ValueError: When trying to learn a model with both nominal and numeric outputs.
    
    Returns:
        list -- List of learned ML-models
    """

    assert isinstance(data, np.ndarray)

    # Init
    if classifier_kwargs is None:
        classifier_kwargs = dict()
    if regressor_kwargs is None:
        regressor_kwargs = dict()

    n_rows, n_cols = data.shape
    attributes = set(range(n_cols))
    nominal_attributes = metadata.get("nominal_attributes")
    numeric_attributes = metadata.get("numeric_attributes")
    msg = """Not all attributes of m_codes are accounted for in metadata"""
    assert not (attributes - nominal_attributes - numeric_attributes), msg

    # Codes to queries
    ids = [
        (d, t)
        for d, t, _ in [code_to_query(m_code, return_list=True) for m_code in m_codes]
    ]

    # Generate a list of random seeds.
    np.random.seed(random_state)
    random_states = np.random.randint(10 ** 3, size=(len(ids)), dtype=int)

    parameters = _build_parameters(
        ids,
        nominal_attributes,
        classifier,
        classifier_kwargs,
        numeric_attributes,
        regressor,
        regressor_kwargs,
        random_states,
        calculation_method_feature_importances,
        min_nb_samples,
        data,
    )

    # Learn Models
    m_list = _build_models(parameters, n_jobs=n_jobs, verbose=verbose)

    return m_list


def expand_induction_algorithm(
    data,
    m_codes,
    metadata,
    classifier,
    regressor,
    classifier_kwargs,
    regressor_kwargs,
    random_state=997,
    n_jobs=1,
    verbose=0,
):
    """Basic induction algorithm. Models according to the m_codes it receives.
    
    Arguments:
        data {np.ndarray} -- Input data
        m_codes {np.ndarray} -- Model codes
        metadata {dict} -- Metadata of MERCS
        classifier {Supported ML learner} -- Supported ML learner
        regressor {Supported ML learner} -- Supported ML learner
        classifier_kwargs {dict} -- Kwargs for classifier
        regressor_kwargs {dict} -- Kwargs for regressor
    
    Keyword Arguments:
        random_state {int} -- Seed for random numbers (default: {997})
        n_jobs {int} -- Joblib can be used in training. (default: {1})
        verbose {int} -- Verbosity level for Joblib. (default: {0})
    
    Raises:
        ValueError: When trying to learn a model with both nominal and numeric outputs.
    
    Returns:
        list -- List of learned ML-models
    """
    m_list = base_induction_algorithm(
        data,
        m_codes,
        metadata,
        classifier,
        regressor,
        classifier_kwargs,
        regressor_kwargs,
        random_state=random_state,
        n_jobs=n_jobs,
        verbose=verbose,
    )

    m_list = _expand_m_list(m_list)

    return m_list


def _expand_m_list(m_list):
    return list(itertools.chain.from_iterable(m_list))


def _build_models(parameters, n_jobs=1, verbose=0):
    m_list = []
    if n_jobs < 2:
        m_list = [_learn_model(*a, **k) for a, k in parameters]
    else:
        msg = """Training is being parallellized using Joblib. Number of jobs = {}""".format(
            n_jobs
        )
        warnings.warn(msg)

        m_list = Parallel(n_jobs=n_jobs, verbose=verbose)(
            delayed(_learn_model)(*a, **k) for a, k in parameters
        )
    return m_list


def _build_parameters(
    ids,
    nominal_attributes,
    classifier,
    classifier_kwargs,
    numeric_attributes,
    regressor,
    regressor_kwargs,
    random_states,
    calculation_method_feature_importances,
    min_nb_samples,
    data,
):
    parameters = []
    for idx, (desc_ids, targ_ids) in enumerate(ids):

        if set(targ_ids).issubset(nominal_attributes):
            learner = classifier
            out_kind = "nominal"
            metric = partial(f1_score, average="macro")
            kwargs = classifier_kwargs.copy()  # Copy is essential
        elif set(targ_ids).issubset(numeric_attributes):
            learner = regressor
            out_kind = "numeric"
            metric = r2_score
            kwargs = regressor_kwargs.copy()  # Copy is essential
        else:
            msg = """
            Cannot learn mixed (i.e. nominal+numeric) models
            """
            raise ValueError(msg)

        kwargs["random_state"] = random_states[idx]
        kwargs[
            "calculation_method_feature_importances"
        ] = calculation_method_feature_importances
        kwargs["min_nb_samples"] = min_nb_samples

        kwargs = _add_categorical_features_to_kwargs(
            learner, desc_ids, nominal_attributes, kwargs
        )  # This is learner-specific

        # Learn a model for current desc_ids-targ_ids combo
        args = (data, desc_ids, targ_ids, learner, out_kind, metric)
        parameters.append((args, kwargs))

    return parameters


def _learn_model(
    data,
    desc_ids,
    targ_ids,
    learner,
    out_kind,
    metric,
    filter_nan=True,
    min_nb_samples=10,
    calculation_method_feature_importances="default",
    **kwargs
):
    """
    Learn a model from the data.

    The arguments of this function determine specifics on which task,
    which learner etc.

    Model is a machine learning method that has a .fit() method.
    """
    assert learner is not None

    i, o = get_i_o(data, desc_ids, targ_ids, filter_nan=filter_nan)

    if i.shape[0] < min_nb_samples:
        msg = """
        Only {} samples available for training.
        min_nb_samples is set to {}.
        Therefore no training occured.
        """.format(
            i.shape[0], min_nb_samples
        )
        warnings.warn(msg)
        return None
    else:
        # Pre-processing
        if i.ndim == 1:
            # We always want 2D inputs
            i = i.reshape(-1, 1)

        multi_target = o.shape[1] != 1
        if not multi_target:
            # If output is single variable, we need 1D matrix
            o = o.ravel()

        if learner in {LGBMC, LGBMR}:
            categorical_feature = kwargs.pop("categorical_feature")
            model = learner(**kwargs)
            model.fit(i, o, categorical_feature=categorical_feature)
        else:
            model = learner(**kwargs)
            model.fit(i, o)

        performance = 1.0

        if calculation_method_feature_importances in {"shap"}:
            model.shap_values_ = _calculate_shap_values(model, i)

        # Bookkeeping
        model = CanonicalModel(model, desc_ids, targ_ids, out_kind, performance)

        return model


# Helpers
def _calculate_shap_values(model, X):
    shap_values = shap.TreeExplainer(model).shap_values(X)
    if isinstance(shap_values, list):
        r = _summarize_shaps(shap_values[0])
    else:
        r = _summarize_shaps(shap_values)
    return r


def _summarize_shaps(shap_values):
    avgs_values = np.mean(np.abs(shap_values), axis=0)
    return np.squeeze(normalize(avgs_values.reshape(1, -1), norm="l1"))


def _add_categorical_features_to_kwargs(learner, desc_ids, nominal_attributes, kwargs):
    assert learner is not None
    cat_features = _get_cat_features(desc_ids, nominal_attributes)

    if learner in {CBC, CBR}:
        kwargs["cat_features"] = cat_features
    elif learner in {WLC, WLR}:
        kwargs["cat_features"] = cat_features
    elif learner in {LGBMC, LGBMR}:
        kwargs["categorical_feature"] = cat_features

    return kwargs


def _score_model(model, metric, y_test, y_pred, multi_target):
    try:
        performance = _calc_performance(y_test, y_pred, metric, multi_target)
    except ValueError as e:
        mean = np.nanmean(y_pred)
        if not np.isfinite(mean):
            warnings.warn(
                """We have a model that cannot solve anything. Something shady might be going on."""
            )
            performance = 0
        else:
            y_pred[np.isnan(y_pred)] = mean
            assert np.all(np.isfinite(y_pred))
            performance = _calc_performance(y_test, y_pred, metric, multi_target)

    return performance


def _calc_performance(y_test, y_pred, metric, multi_target):
    if multi_target:
        performance = [
            metric(y_test[:, i], y_pred[:, i]) for i in range(y_test.shape[1])
        ]
    else:
        performance = metric(y_test, y_pred)
    return performance


def _get_cat_features(desc_ids, nominal_ids):
    cat_features = np.where(
        np.in1d(np.array(list(desc_ids)), np.array(list(nominal_ids)))
    )[0]
    return [int(c) for c in cat_features]



import networkx as nx
import numpy as np

from itertools import product
from .graphviz import fix_layout


NODE_PREFIXES = dict(
    data="d",
    vote="v",
    prob="p",
    merge="M",
    imputation="I",
    imputer="I",
    function="f",
    composition="c",
    model="f")

SHAPES = dict(
    data='"circle',
    vote='"triangle"',
    prob='"circle',
    merge='"triangle"',
    imputation='"invtriangle"',
    imputer='"invtriangle"',
    function='"square"',
    composition='"square"',
    model='"square"')


# Building a graph
def model_to_graph(model, types=None, idx=0, composition=False):
    """
    Convert a model to a network.

    Given a model with inputs and outputs, we convert it to a graphical
    model that represents it.

    Parameters
    ----------
    types
    model:      model
                Model which is to be converted to a schematic representation
    idx:        int, default=0
                index (or equivalent:id) of the model

    Returns
    -------

    TODO
    ----
    Numpy throws a runtimewarning somewhere in this method, I need to fix that.

    """

    G = nx.DiGraph()

    # Create nodes

    if composition:
        func_nodes = [_create_comp_node(model, idx=idx)]
    else:
        func_nodes = [_create_func_node(model, idx=idx)]

    nominal = func_nodes[0][1]["out"] == "nominal"  # Nominal model yes/no

    data_nodes_src = [_create_data_node(i, types) for i in model.desc_ids]
    data_nodes_tgt = [_create_data_node(i, types) for i in model.targ_ids]

    G.add_nodes_from(data_nodes_src)
    G.add_nodes_from(data_nodes_tgt)
    G.add_nodes_from(func_nodes)

    if nominal:
        classes = get_classes(model)
        prob_nodes_tgt = [
            _create_prob_node(i, types, classes) for i in model.targ_ids
        ]
        vote_nodes_tgt = [_create_vote_node(i, types, classes) for i in model.targ_ids]
        G.add_nodes_from(prob_nodes_tgt)
        G.add_nodes_from(vote_nodes_tgt)

    # Create edges
    f = [t[0] for t in func_nodes]
    s = [t[0] for t in data_nodes_src]
    t = [t[0] for t in data_nodes_tgt]

    src_to_fnc_edges = [
        (*e, {"idx": d, "fi": fi})
        for e, d, fi in zip(product(s, f), model.desc_ids, model.feature_importances_)
    ]
    G.add_edges_from(src_to_fnc_edges)

    if nominal:
        p = [t[0] for t in prob_nodes_tgt]
        v = [t[0] for t in vote_nodes_tgt]

        fnc_to_prb_edges = [
            (*e, {"idx": d, "classes": G.node[e[1]]["classes"]})
            for e, d in zip(product(f, p), model.targ_ids)
        ]
        prb_to_vot_edges = [
            (*e, {"idx": d}) for e, d in zip(product(p, v), model.targ_ids)
        ]
        vot_to_tgt_edges = [
            (*e, {"idx": d}) for e, d in zip(product(v, t), model.targ_ids)
        ]

        G.add_edges_from(fnc_to_prb_edges)
        G.add_edges_from(prb_to_vot_edges)
        G.add_edges_from(vot_to_tgt_edges)
    else:
        mod_to_tgt_edges = [
            (*e, {"idx": d}) for e, d in zip(product(f, t), model.targ_ids)
        ]
        G.add_edges_from(mod_to_tgt_edges)

    # Graph attributes
    # TODO: Assign this in a single place. In the core file, we assign this to the composites.
    G.graph["desc_ids"] = set(model.desc_ids)
    G.graph["targ_ids"] = set(model.targ_ids)
    G.graph["id"] = idx

    # FI
    G = add_fi_to_graph(G)

    # Final touch
    fix_layout(G)
    return G


def _create_comp_node(model, idx=0):
    comp_node = (
        node_label(idx, kind="composition"),
        dict(
            kind="model",
            idx=idx,
            mod=model,
            src=model.desc_ids,
            tgt=model.targ_ids,
            out=model.out_kind
        ),
    )

    if comp_node[1]["out"] in {"numeric"}:
        comp_node[1]["predict"] = model.predict
        
        # No need to do inference, this goes right away
        comp_node[1]["dask"] = comp_node[1]["predict"]
    elif comp_node[1]["out"] in {"nominal", "mix"}:
        comp_node[1]["predict"] = model.predict
        comp_node[1]["predict_proba"] = model.predict_proba

        # No need to do inference, this goes right away
        comp_node[1]["dask"] = comp_node[1]["predict"]
        comp_node[1]["dask_proba"] = comp_node[1]["predict_proba"]
    else:
        pass
    return comp_node


def _create_func_node(model, idx=0):

    func_node = (
        node_label(idx, kind="function"),
        dict(
            kind="model",
            idx=idx,
            mod=model,
            src=model.desc_ids,
            tgt=model.targ_ids,
            out=model.out_kind
        ),
    )

    if func_node[1]["out"] in {"numeric"}:
        func_node[1]["predict"] = model.predict
    elif func_node[1]["out"] in {"nominal", "mix"}:
        func_node[1]["predict"] = model.predict
        func_node[1]["predict_proba"] = model.predict_proba
    else:
        pass
    return func_node


def _create_data_node(idx, types):
    data_node = (
        node_label(idx, kind="data"),
        dict(kind="data", idx=idx, tgt=[idx], type=types[idx]),
    )
    return data_node


def _create_prob_node(idx, types, classes):
    prob_node = (
        node_label(idx, kind="prob"),
        dict(kind="prob", idx=idx, tgt=[idx], type=types[idx], classes=classes[idx]),
    )
    return prob_node


def _create_vote_node(idx, types, classes):
    vote_node = (
        node_label(idx, kind="vote"),
        dict(kind="vote", idx=idx, tgt=[idx], type=types[idx], classes=classes[idx]),
    )


    return vote_node


def compose_all(g_list):
    R = g_list[0].__class__()

    # Add nodes
    classes = {}
    for g in g_list:
        # Add nodes (one at a time)
        for n, d in g.nodes(data=True):
            R.add_node(n, **d)

            if d["kind"] in {"vote", "prob"}:
                n_classes = d.get("classes", np.array([]))
                r_classes = classes.get(n, np.array([]))
                classes[n] = np.concatenate([r_classes, n_classes])

        # Add edges (in bulk)
        R.add_edges_from(g.edges(data=True))

    # Joining classes
    all_prob_nodes = get_nodes(R, kind="prob")
    all_vote_nodes = get_nodes(R, kind="vote")
    for n in all_prob_nodes.union(all_vote_nodes):
        R.node[n]["classes"] = np.unique(classes[n])

    return R


# Add FI
def add_fi_to_graph(G):
    """
    Add feature importances to input data nodes.

    Parameters
    ----------
    G

    Returns
    -------

    """

    src_nodes = (
        n
        for n, in_degree in G.in_degree()
        if G.nodes[n]["kind"] == "data"
        if in_degree == 0
    )

    src_edges = ((n, G.out_edges(n)) for n in src_nodes)

    for n, edges in src_edges:
        fi = np.mean([G.edges()[e]["fi"] for e in edges])
        G.nodes[n]["fi"] = fi

    return G


# Add
def add_merge_nodes(g):

    relevant_nodes = [n for n in get_nodes(g, kind="data") if len(g.in_edges(n)) > 1]

    for node in relevant_nodes:
        # Collect original node
        original_node_attributes = g.nodes(data=True)[node]
        original_node = (node, original_node_attributes)

        # Convert original
        merge_node_label = convert_data_node(g, node, kind="merge")

        # Insert the original back in, after the merge node
        insert_new_node(g, original_node, merge_node_label, location="back")
    return


def add_imputation_nodes(g, q_desc):

    relevant_nodes = [n for n in get_desc(g, ids=False)
                      if g.nodes[n]["idx"] not in q_desc]

    for node in relevant_nodes:
        convert_data_node(g, node, kind="imputation")
    return


# Node operations
def convert_data_node(G, data_node_label, kind="merge"):

    # Init
    mapping = {}
    prefix = NODE_PREFIXES[kind]
    shape = SHAPES[kind]

    # Actions
    mapping[data_node_label] = "{}({})".format(prefix, data_node_label)
    new_node_label = mapping[data_node_label]

    nx.relabel_nodes(G, mapping, copy=False)

    G.nodes[new_node_label]["shape"] = shape
    G.nodes[new_node_label]["kind"] = kind

    return mapping[data_node_label]


def insert_new_node(G, new_node, existing_node_label, location="behind"):
    G.add_node(new_node[0], **new_node[1])

    if location in {"behind", "after", "back"}:
        G.add_edge(existing_node_label, new_node[0], idx=new_node[1]["idx"])
    elif location in {"before", "front"}:
        G.add_edge(new_node[0], existing_node_label, idx=new_node[1]["idx"])
    else:
        msg = """
        Did not recognize location:     {}
        to add new node relative to existing node.
        """.format(
            location
        )
        raise ValueError(msg)

    return


def compose(G, H):
    """
    Return a new graph of G composed with H.

    Composition is the simple union of the node sets and edge sets.
    The node sets of G and H need not be disjoint.

    Parameters
    ----------
    G, H : graph
       A NetworkX graph

    Returns
    -------
    C: A new graph  with the same type as G

    Notes
    -----
    This is a custom version of nx.compose
    """

    if not G.is_multigraph() == H.is_multigraph():
        raise nx.NetworkXError("G and H must both be graphs or multigraphs.")

    R = G.__class__()

    # add graph attributes, H attributes take precedent over G attributes
    R.graph.update(G.graph)
    R.graph.update(H.graph)

    R.add_nodes_from(G.nodes(data=True))
    R.add_nodes_from(H.nodes(data=True))

    if G.is_multigraph():
        R.add_edges_from(G.edges(keys=True, data=True))
    else:
        R.add_edges_from(G.edges(data=True))
    if H.is_multigraph():
        R.add_edges_from(H.edges(keys=True, data=True))
    else:
        R.add_edges_from(H.edges(data=True))

    # Custom modification
    all_prob_nodes = get_nodes(R, kind="prob")
    all_vote_nodes = get_nodes(R, kind="vote")
    for node in all_prob_nodes.union(all_vote_nodes):
        a = get_attribute(G, node, "classes", np.array([]))
        b = get_attribute(H, node, "classes", np.array([]))

        R.node[node]["classes"] = np.unique(np.concatenate([a, b]))
    return R


# Utils
def get_attribute(G, node, attribute, default=None):
    """
    From a node that may not exist, collect an attribute that may not be there.

    Parameters
    ----------
    G
    node
    attribute
    default

    Returns
    -------

    """

    return G.nodes.get(node, {}).get(attribute, default)


def get_classes(model):
    if len(model.targ_ids) > 1:
        return {
            tgt_idx: model.classes_[idx] for idx, tgt_idx in enumerate(model.targ_ids)
        }
    else:
        # Single target has no indexes, so just pass model.classes_ directly
        return {tgt_idx: model.classes_ for idx, tgt_idx in enumerate(model.targ_ids)}


def node_label(idx, kind="function"):
    """
    Generate a unique name for a node.

    Args:
        idx:        int
                    Node id
        kind:       str, {'func', 'model', 'function'} or {'data}
                    Every node represents either a function or data.

    Returns:

    """
    prefix = NODE_PREFIXES[kind]
    return "{}-{:02d}".format(prefix, idx)


def get_ids(g, kind="desc"):

    if kind in {"s", "src", "source", "d", "desc", "descriptive"}:
        r = get_desc(g, ids=True)
    elif kind in {"t", "tgt", "targ", "target"}:
        r = get_targ(g, ids=True)
    else:
        msg = """
        Did not recognize kind:   {}
        """.format(
            kind
        )
        raise ValueError(msg)

    return r


def get_desc(g, ids=False):
    data_nodes = get_nodes(g, kind="data")

    r = {n for n in data_nodes if len(g.in_edges(n)) == 0}

    if ids:
        r = {g.nodes[n]["idx"] for n in r}

    return r


def get_targ(g, ids=False):
    data_nodes = get_nodes(g, kind="data")

    r = {n for n in data_nodes if len(g.out_edges(n)) == 0}

    if ids:
        r = {g.nodes[n]["idx"] for n in r}

    return r


def get_nodes(g, kind="data"):
    prefix = NODE_PREFIXES[kind]
    r = {n for n in g.nodes if n.startswith(prefix)}
    return r



import copy
import warnings
from functools import reduce

import networkx as nx
import numpy as np

from .new_prediction import _stochastic_pick
from ..graph import add_imputation_nodes, add_merge_nodes, compose, get_ids, get_nodes
from ..utils import code_to_query, debug_print, encode_attribute, query_to_code

DESC_ENCODING = encode_attribute(1, [1], [2])
TARG_ENCODING = encode_attribute(2, [1], [2])
MISS_ENCODING = encode_attribute(0, [1], [2])

VERBOSITY = 0


def mi_algorithm(g_list, q_code, random_state=997):
    q_desc, q_targ, q_miss = code_to_query(q_code)

    def criterion(g):
        """If g predicts a relevant target = OK"""
        tgt_ids = get_ids(g, kind="tgt")
        return len(set(q_targ).intersection(tgt_ids)) > 0

    g_relevant = [g for g in g_list if criterion(g)]
    g_relevant = [copy.copy(g) for g in g_relevant]

    result = reduce(compose, g_relevant)

    # Post-processing to get a good graph
    add_imputation_nodes(result, q_desc)
    add_merge_nodes(result)

    return result


def ma_algorithm(g_list, q_code, init_threshold=1.00, stepsize=0.05, random_state=997):
    q_desc, q_targ, q_miss = code_to_query(q_code)

    def criterion(g):
        src_ids = get_ids(g, kind="src")
        tgt_ids = get_ids(g, kind="tgt")

        quantifier = len(set(q_desc).intersection(src_ids)) / len(src_ids)
        rel_tgt = tgt_ids.intersection(q_targ)
        yes_no = len(rel_tgt) > 0
        relevance_criterion = int(yes_no) * quantifier

        msg = """
        rel_tgt:                     {}
        quantifier:                 {}
        relevance_criterion:        {}
        """.format(
            rel_tgt, quantifier, relevance_criterion
        )
        debug_print(msg, level=1, V=VERBOSITY)

        return relevance_criterion

    thresholds = np.clip(np.arange(init_threshold, -stepsize, -stepsize), 0, 1)

    for thr in thresholds:
        g_relevant = [g for g in g_list if criterion(g) > thr]
        if len(g_relevant) > 0:
            msg = """
            We have selected {0} model(s) at threshold: {1:.2f}
            """.format(
                len(g_relevant), thr
            )
            debug_print(msg, level=1, V=VERBOSITY)
            break

    g_relevant = [copy.copy(g) for g in g_relevant]
    result = reduce(compose, g_relevant)

    # Post-processing to get a good graph
    add_imputation_nodes(result, q_desc)
    add_merge_nodes(result)

    return result


def mrai_algorithm(
    g_list,
    q_code,
    init_threshold=1.0,
    stepsize=0.1,
    avoid_src=None,
    avoid_tgt=None,
    return_avl_g=False,
    greedy=True,
    stochastic=False,
    imputation_nodes=True,
    merge_nodes=True,
    random_state=997,
):

    # Preliminaries
    if avoid_src is None:
        avoid_src = set([])
    if avoid_tgt is None:
        avoid_tgt = set([])

    q_desc, q_targ, q_miss = code_to_query(q_code)
    nb_tgt = len(q_targ)

    thresholds = np.arange(init_threshold, -1 - stepsize, -stepsize)
    thresholds = np.clip(thresholds, -1, 1)

    # Methods
    def stopping_criterion(list_of_graphs):
        return len(list_of_graphs) > 0

    def criterion(g):
        src_ids = get_ids(g, kind="src")
        tgt_ids = get_ids(g, kind="tgt")

        rel_src = src_ids.intersection(q_desc)
        rel_tgt = tgt_ids.intersection(q_targ)

        bad_src = src_ids.intersection(set(q_targ).union(avoid_src))
        bad_tgt = tgt_ids.intersection(avoid_tgt)  # Things I should not predict

        data_nodes = get_nodes(g, kind="data")

        quantifier = 0
        for n in data_nodes:
            idx = g.node[n]["idx"]

            if idx in rel_src:
                quantifier += g.node[n]["fi"]
            if idx in bad_src:
                quantifier -= g.node[n]["fi"]

        f_01 = max(0.1, quantifier)
        f_02 = int(len(rel_tgt) > 0)
        f_03 = int(len(bad_tgt) > 0)

        if f_02:
            return f_01
        elif f_03:
            # print("Bad target: {}".format(bad_tgt))
            # No relevant targets, and a bad target. This model contributes nothing.
            return -1
        else:
            return -1

    # Actual algorithm
    if stochastic:
        criteria = [criterion(g) for g in g_list]
        picks = _pick(criteria, n=1, random_state=random_state)

        msg = """
                criteria:   {}
                picks:      {}
                """.format(
            criteria, picks
        )
        debug_print(msg, level=1, V=VERBOSITY)

        sel_g = [g_list[g_idx] for g_idx in picks]

        sel_g = [copy.deepcopy(g) for g in sel_g]
        res_g = reduce(compose, sel_g)

    elif nb_tgt == 1 or greedy:
        # Don't stop until you reach the goal
        criteria = [criterion(g) for g in g_list]

        for thr in thresholds:
            sel_g = [
                g_list[g_idx] for g_idx, c in enumerate(criteria) if c > thr
            ]  # Available graphs = Graphs that satisfy the criterion

            if stopping_criterion(sel_g):
                mod_ids = [get_nodes(g, kind="model") for g in sel_g]
                msg = """
                We have selected    {0} model(s) 
                at threshold:       {1:.2f}
                with model ids:     {2}
                """.format(
                    len(sel_g), thr, mod_ids
                )
                debug_print(msg, level=1, V=VERBOSITY)
                break

        sel_g = [copy.deepcopy(g) for g in sel_g]

        for g in sel_g:
            add_imputation_nodes(g, q_desc)

        res_g = reduce(compose, sel_g)

    elif nb_tgt > 1:
        msg = """
        Multi-target case:      
        with target attributes: {}
        """.format(
            q_targ
        )
        debug_print(msg, V=VERBOSITY)

        sel_g = []
        avl_g = g_list
        for i in range(len(q_targ)):
            tgt_q_targ = q_targ[i : i + 1]
            tgt_avoid_src = q_targ[0:i] + q_targ[i + 1 :]
            tgt_q_miss = q_miss + tgt_avoid_src

            tgt_q_code = query_to_code(q_desc, tgt_q_targ, q_miss=tgt_q_miss)

            msg = """
            tgt_q_code:     {}
            """.format(
                tgt_q_code
            )
            debug_print(msg, level=2, V=VERBOSITY)
            tgt_g, avl_g = mrai_algorithm(
                avl_g,
                tgt_q_code,
                avoid_src=tgt_avoid_src,
                return_avl_g=True,
                merge_nodes=False,
            )

            sel_g.append(tgt_g)

        res_g = reduce(compose, sel_g)

    else:
        msg = """
        nb_tgt:     {}
        We expect one or more targets.
        """.format(
            nb_tgt
        )
        raise ValueError(msg)

    if imputation_nodes:
        add_imputation_nodes(res_g, q_desc)
    if merge_nodes:
        add_merge_nodes(res_g)

    if return_avl_g:
        avl_g = [g for g in g_list if g not in sel_g]
        return res_g, avl_g
    else:
        return res_g


def it_algorithm(g_list, q_code, max_steps=4, random_state=997):
    """

    Notes
    -----
    `avl` is short for `available`

    avl_desc:   Available descriptive attributes = Attributes that are available AS descriptive ones.
                So, this means these attributes are KNOWN.
    avl_targ:   Available target attributes = Attributes that are available AS target ones.
                So, this means these attributes are UNKNOWN.

    Parameters
    ----------
    g_list
    q_code
    max_steps
    random_state

    Returns
    -------

    """

    def stopping_criterion(known_attributes, target_attributes):
        """All targets known = DONE"""
        return len(set(target_attributes).difference(known_attributes)) == 0

    # Init
    q_desc, q_targ, q_miss = code_to_query(q_code, return_list=True)
    avl_desc = set(q_desc)
    avl_targ = set(q_targ + q_miss)
    avl_atts = set(q_desc + q_targ + q_miss)

    avl_q = query_to_code(avl_desc, avl_targ, atts=avl_atts)  # 0: Known, 1: Unknown

    avl_g = g_list

    greedy = True
    sel_g = []
    for step in range(max_steps):
        last = step == (max_steps - 1)  # Check if this is our last chance

        if last:
            # All unknown targets are the target of the last step.
            avl_targ = set(q_targ).difference(avl_desc)
            avl_q = query_to_code(avl_desc, avl_targ, atts=avl_atts)
            greedy = False  # Get all remaining targets

        # Get next step
        nxt_g, avl_g = mrai_algorithm(
            avl_g,
            avl_q,
            return_avl_g=True,
            greedy=greedy,
            avoid_src=q_targ,
            avoid_tgt=avl_desc,
            imputation_nodes=True,
            merge_nodes=False,
            random_state=random_state,
        )
        # IT goes from front to back
        sel_g.append(nxt_g)

        # Update query
        nxt_targ = get_ids(nxt_g, kind="targ")
        avl_desc = avl_desc.union(nxt_targ)
        avl_targ = avl_targ.difference(nxt_targ)

        avl_q = query_to_code(avl_desc, avl_targ, atts=avl_atts)

        if stopping_criterion(avl_desc, q_targ):
            break

    # Composing
    res_g = nx.DiGraph()
    avl_desc = set(q_desc)
    for g in sel_g:
        msg = """
        AVL DESC: {}
        """.format(
            avl_desc
        )
        debug_print(msg, level=1, V=VERBOSITY)

        # add_imputation_nodes(g, avl_desc)
        res_g = compose(res_g, g)

        g_targ = get_ids(g, kind="targ")
        avl_desc = avl_desc.union(g_targ)

    add_merge_nodes(res_g)
    res_g = _prune(res_g, q_targ)

    return res_g


def rw_algorithm(g_list, q_code, max_steps=4, random_state=997):
    def stopping_criterion(targets, step_number):
        reason_01 = len(targets) == 0
        reason_02 = step_number == max_steps - 1
        return reason_01 or reason_02

    # Init - diagrams
    avl_g = g_list
    sel_g = []

    # Init - Attribute sets
    q_desc, q_targ, q_miss = code_to_query(q_code, return_list=True)

    sel_targ = set([])
    avl_desc = set(q_desc)
    avl_targ = set(q_targ)
    avl_atts = set(q_desc + q_targ + q_miss)
    avl_q = q_code

    for step in range(max_steps):
        # Get next step
        nxt_g, avl_g = mrai_algorithm(
            avl_g,
            avl_q,
            return_avl_g=True,
            avoid_src=q_targ,
            avoid_tgt=avl_desc,
            stochastic=True,
            merge_nodes=False,
            imputation_nodes=False,
            random_state=random_state,
        )
        # RW goes from back to front
        sel_g.insert(0, nxt_g)

        # Update query
        nxt_g_desc = get_ids(nxt_g, kind="desc")
        nxt_g_targ = get_ids(nxt_g, kind="targ")

        sel_targ = sel_targ.union(nxt_g_targ)

        avl_desc = avl_desc
        avl_targ = set(q_miss).difference(sel_targ).intersection(nxt_g_desc)
        avl_q = query_to_code(avl_desc, avl_targ, atts=avl_atts)

        if stopping_criterion(avl_targ, step):
            break

    # Composing
    res_g = nx.DiGraph()
    avl_desc = set(q_desc)
    for g in sel_g:
        msg = """
        AVL DESC: {}
        """.format(
            avl_desc
        )
        debug_print(msg, level=1, V=VERBOSITY)
        add_imputation_nodes(g, avl_desc)
        res_g = compose(res_g, g)

        g_targ = get_ids(g, kind="targ")
        avl_desc = avl_desc.union(g_targ)

    add_merge_nodes(res_g)
    res_g = _prune(res_g, q_targ)

    return res_g


# Helpers
def _prune(g, tgt_nodes=None):

    msg = """
    tgt_nodes:          {}
    tgt_nodes[0]:       {}
    type(tgt_nodes[0]): {}
    """.format(
        tgt_nodes, tgt_nodes[0], type(tgt_nodes[0])
    )
    debug_print(msg, level=1, V=VERBOSITY)

    if tgt_nodes is None:
        tgt_nodes = [
            n
            for n, out_degree in g.out_degree()
            if out_degree == 0
            if g.nodes[n]["kind"] == "data"
        ]
        msg = """
        tgt_nodes: {}
        """.format(
            tgt_nodes
        )
        debug_print(msg, level=1, V=VERBOSITY)
    elif isinstance(tgt_nodes[0], (int, np.int64)):
        tgt_nodes = [
            n
            for n in g.nodes
            if g.nodes[n]["kind"] == "data"
            if g.nodes[n]["idx"] in tgt_nodes
        ]
    else:
        assert isinstance(tgt_nodes[0], str)

    ancestors = [nx.ancestors(g, source=n) for n in tgt_nodes]
    retain_nodes = reduce(set.union, ancestors, set(tgt_nodes))

    nodes_to_remove = [n for n in g.nodes if n not in retain_nodes]
    for n in nodes_to_remove:
        g.remove_node(n)

    return g

def _pick(criteria, n=1, random_state=997):
    """
    Interpret an array of appropriateness scores as a distribution
    corresponding to the probability of a certain model being selected.


    Parameters
    ----------
    criteria:   list
                Array that quantifies how likely a pick should be.
    n:          int
                Number of picks

    Returns
    -------
    picks:      np.ndarray
                List of indices that were picked
    """

    np.random.seed(random_state)
    criteria += abs(np.min([0, np.min(criteria)]))  # Shift in case of negative values
    norm = np.linalg.norm(criteria, 1)

    if norm > 0:
        criteria = criteria / norm
    else:
        msg = """
        Not a single appropriate model was found, therefore
        making an arbitrary choice.
        """
        warnings.warn(msg)
        # If you cannot be right, be arbitrary
        criteria = [1 / len(criteria) for i in criteria]

    draw = np.random.multinomial(1, criteria, size=n)
    picks = np.where(draw == 1)[1]
    return picks


from functools import lru_cache

def o(f1, f2):

    def sequential_composition(*function_arguments):
        return f1(f2(*function_arguments))

    return sequential_composition


def x(*functions, return_type=tuple):
    def parallel_composition(*function_arguments):
        if len(function_arguments) == 1:
            res = return_type([f(function_arguments[0]) for f in functions])
            return res
        elif len(function_arguments) == len(functions):
            res = return_type([f(x) for f, x in zip(functions, function_arguments)])
            return res

    return parallel_composition



import numpy as np


def encode_attribute(att, desc, targ):
    """
    Encode the 'role' of an attribute in a model.

    This is an important function, since here resides the precise encoding.
    If this is changed, a whole lot of other things will change too. Since everything
    that relies on the encoding should refer here, everything should proceed normally.

    `Role` means:
        - Descriptive attribute (input)
        - Target attribute (output)
        - Missing attribute (not relevant to the model)
    """

    check_desc = att in desc
    check_targ = att in targ

    code_int = check_targ * 2 + check_desc - 1

    return code_int


# Ugly but it just needs to be done once so deal with it.
DESC_ENCODING = encode_attribute(1, [1], [2])
TARG_ENCODING = encode_attribute(2, [1], [2])
MISS_ENCODING = encode_attribute(0, [1], [2])

ENCODING = dict(desc=DESC_ENCODING, targ=TARG_ENCODING, miss=MISS_ENCODING)


def code_to_query(code, return_sets=False, return_list=False):

    desc = get_att(code, kind="desc")
    targ = get_att(code, kind="targ")
    miss = get_att(code, kind="miss")

    if return_sets:
        return set(desc), set(targ), set(miss)
    elif return_list:
        return list(desc), list(targ), list(miss)
    else:
        return desc, targ, miss


def get_att(code, kind="desc"):
    return np.where(code == ENCODING[kind])[0].astype(int)


def get_att_2d(codes, kind="desc"):
    return np.unique(np.where(codes == ENCODING[kind])[1].astype(int))


def query_to_code(q_desc, q_targ, q_miss=None, attributes=None):

    if attributes is None:
        attributes = determine_attributes(q_desc, q_targ, q_miss)

    code = [encode_attribute(a, q_desc, q_targ) for a in attributes]

    return np.array(code)


def change_role(
    q_code, source=ENCODING["miss"], target=ENCODING["targ"], inplace=False
):
    if not inplace:
        q_code = q_code.copy()

    q_code[q_code == source] = target
    return q_code


def determine_attributes(desc, targ, miss=None):
    """
    Determine the entire list of attributes.
    """
    if miss is None:
        miss = []

    atts = list(set(desc).union(targ).union(miss))
    atts.sort()
    return atts



import numpy as np
import warnings

from sklearn.preprocessing import minmax_scale, normalize

from ..utils import code_to_query, get_att_2d, TARG_ENCODING

EPSILON = 0.00001


# Strategies
def mi(
    m_codes,
    m_fimps,
    m_score,
    q_code,
    a_src=None,
    a_tgt=None,
    m_avl=None,
    random_state=997,
):
    # Init
    m_sel = []
    m_avl = _init_m_avl(m_codes, m_avl=m_avl)
    a_src, a_tgt = _init_a_src_a_tgt(q_code=q_code, a_src=a_src, a_tgt=a_tgt)

    c_all = criterion(m_codes, row_filter=m_avl, col_filter=a_tgt, aggregation=False)
    m_sel_idx = np.unique(np.where(c_all == TARG_ENCODING)[0]).astype(int)
    m_sel = m_avl[m_sel_idx]

    if len(m_sel) == 0:
        return None
    else:
        return m_sel


def mrai(
    m_codes,
    m_fimps,
    m_score,
    q_code,
    a_src=None,
    a_tgt=None,
    m_avl=None,
    init_threshold=1.0,
    stepsize=0.1,
    any_target=False,
    picking_function="greedy",
    thresholds=None,
    random_state=997,
):
    # Init
    m_sel = []
    m_avl = _init_m_avl(m_codes, m_avl=m_avl)
    thresholds = _init_thresholds(init_threshold, stepsize)
    a_src, a_tgt = _init_a_src_a_tgt(q_code=q_code, a_src=a_src, a_tgt=a_tgt)

    # Filtering
    m_flt = mi(
        m_codes,
        m_fimps,
        m_score,
        None,
        a_src=a_src,
        a_tgt=a_tgt,
        m_avl=m_avl,
        random_state=random_state,
    )

    if m_flt is None:
        warnings.warn("You reached a dead end.")
        return None
    else:
        # Criterion
        c_tgt = criterion(
            m_score, row_filter=m_flt, col_filter=a_tgt, aggregation=False
        )
        c_src = criterion(m_fimps, row_filter=m_flt, col_filter=a_src, aggregation=True)
        c_all = c_src * c_tgt + EPSILON

        if any_target:
            c_all = np.max(c_all, axis=1).reshape(-1, 1)

        # Normalize
        c_nrm = normalize(c_all, norm="l1", axis=0)
        c_nrm += 1 - np.max(c_nrm)

        # Pick
        m_sel_idx = pick(
            c_all,
            thresholds=thresholds,
            picking_function=picking_function,
            random_state=random_state,
        )
        m_sel = m_flt[m_sel_idx]

        return m_sel


def it(
    m_codes,
    m_fimps,
    m_score,
    q_code,
    m_avl=None,
    max_steps=4,
    init_threshold=1.0,
    stepsize=0.1,
    picking_function="greedy",
    random_state=997,
):
    m_sel = []
    m_avl = _init_m_avl(m_codes, m_avl=m_avl)
    thresholds = _init_thresholds(init_threshold, stepsize)

    any_target = True
    q_desc, q_targ, q_miss = code_to_query(q_code)
    a_src = q_desc
    a_tgt = np.hstack([q_targ, q_miss])

    for step in range(max_steps):
        # Check if this is our last chance
        last = step + 1 == max_steps
        if last:
            any_target = False  # Finish the job
            a_tgt = np.setdiff1d(
                q_targ, a_src
            )  # Focus exclusively on non-predicted q_targ attributes.

        step_m_sel = mrai(
            m_codes,
            m_fimps,
            m_score,
            None,
            a_src=a_src,
            a_tgt=a_tgt,
            m_avl=m_avl,
            any_target=any_target,
            thresholds=thresholds,
            picking_function=picking_function,
            random_state=random_state,
        )

        if step_m_sel is None:
            raise ValueError(
                "No progress was made. This indicates an impossible query."
            )

        a_prd = get_att_2d(m_codes[step_m_sel, :], kind="targ")

        a_src = np.union1d(a_src, a_prd)
        a_tgt = np.setdiff1d(a_tgt, a_prd)

        m_avl = np.setdiff1d(m_avl, step_m_sel)
        m_sel.append(step_m_sel)

        if _stopping_criterion_it(q_targ, a_src):
            break

    return m_sel


def rw(
    m_codes,
    m_fimps,
    m_score,
    q_code,
    m_avl=None,
    nb_walks=5,
    max_steps=4,
    init_threshold=1.0,
    stepsize=0.1,
    straight=True,
    picking_function="stochastic",
    random_state=997,
):

    random_walks = [
        walk(
            m_codes,
            m_fimps,
            m_score,
            q_code,
            m_avl=m_avl,
            max_steps=max_steps,
            init_threshold=init_threshold,
            stepsize=stepsize,
            straight=straight,
            picking_function=picking_function,
            random_state=random_state + i * 997,  # Otherwise you do identical walks!
        )
        for i in range(nb_walks)
    ]

    return tuple(random_walks)


def walk(
    m_codes,
    m_fimps,
    m_score,
    q_code,
    m_avl=None,
    max_steps=4,
    init_threshold=1.0,
    stepsize=0.1,
    straight=True,
    picking_function="stochastic",
    random_state=997,
):
    m_sel = []
    m_avl = _init_m_avl(m_codes, m_avl=m_avl)
    thresholds = _init_thresholds(init_threshold, stepsize)

    any_target = True
    q_desc, q_targ, q_miss = code_to_query(q_code)
    a_src = q_desc
    a_tgt = q_targ

    for step in reversed(range(max_steps)):

        step_m_sel = mrai(
            m_codes,
            m_fimps,
            m_score,
            None,
            a_src=a_src,
            a_tgt=a_tgt,
            m_avl=m_avl,
            any_target=any_target,
            thresholds=thresholds,
            picking_function=picking_function,
            random_state=random_state + step,
        )

        if step_m_sel is None:
            break

        # Step source attributes = potential targets
        s_src = get_att_2d(m_codes[step_m_sel, :], kind="desc")

        # Allow to predict anything that follows
        if not straight:
            s_tgt = get_att_2d(m_codes[step_m_sel, :], kind="targ")  # HACK
            s_src = np.union1d(a_tgt, s_src)  # HACK

        a_tgt = np.setdiff1d(s_src, a_src)

        m_avl = np.setdiff1d(m_avl, step_m_sel)
        m_sel.insert(0, step_m_sel)

        if _stopping_criterion_rw(a_tgt, step, m_avl):
            break

    return m_sel


# MRAI-IT-RW Criterion
def criterion(matrix, row_filter=None, col_filter=None, aggregation=False):
    """
    Typical usecase 
    
    matrix = m_fimps
    
    row_filter = available models
    col_filter = available attributes.
    
    """
    nb_rows = len(row_filter)
    nb_cols = len(col_filter)

    m_idx = col_filter + row_filter.reshape(-1, 1) * matrix.shape[1]
    c_matrix = matrix.take(m_idx.flat).reshape(nb_rows, nb_cols)

    if aggregation:
        return np.sum(c_matrix, axis=1).reshape(-1, 1)
    else:
        return c_matrix


# Picks
def _all_pick(criterion, **kwargs):
    return np.where(criterion >= 0.0)[0]


def _greedy_pick(criterion, thresholds=None, **kwargs):

    for thr in thresholds:
        m_sel = np.where(criterion >= thr)[0]

        if _stopping_criterion_greedy_pick(m_sel):
            break
    return m_sel


def _stochastic_pick(c_all, random_state=997, normalize=True, **kwargs):
    assert len(c_all) > 0

    if normalize:
        c_all = _criteria_to_distribution(c_all)

    np.random.seed(random_state)

    try:
        draw = np.random.multinomial(1, c_all, size=1)
    except ValueError:
        draw = np.random.multinomial(1, c_all * 0.95, size=1)

    m_sel = np.where(draw == 1)[1]
    return m_sel


def _random_pick(criterion, random_state=997, **kwargs):
    criterion = _criteria_to_uniform_distribution(criterion)
    return _stochastic_pick(criterion, normalize=False)


PICKS = dict(
    all=_all_pick, greedy=_greedy_pick, stochastic=_stochastic_pick, random=_random_pick
)


def pick(criteria, thresholds=None, picking_function="greedy", random_state=997):
    m_sel = []

    for c_idx in range(criteria.shape[1]):
        c_act = criteria[:, c_idx]
        if not np.any(c_act):
            c_act = _criteria_to_uniform_distribution(c_act)

        m_sel.extend(
            PICKS[picking_function](
                c_act, thresholds=thresholds, random_state=random_state
            )
        )

    return np.unique(m_sel)


# Stopping Criteria
def _stopping_criterion_rw(q_targ, step, m_avl):
    reason_01 = len(q_targ) == 0
    reason_02 = step == 0
    reason_03 = len(m_avl) == 0
    return reason_01 or reason_02 or reason_03


def _stopping_criterion_it(q_targ, a_src):
    return np.setdiff1d(q_targ, a_src).shape[0] == 0


def _stopping_criterion_greedy_pick(m_sel):
    return len(m_sel) > 0


# Helpers
def _criteria_to_uniform_distribution(criteria):
    return np.full(criteria.shape, 1 / criteria.shape[0])


def _criteria_to_distribution(criteria, epsilon=EPSILON):
    # Criteria = VECTOR in [0,1]
    norm = np.sum(criteria)

    if norm < epsilon:
        dist = _criteria_to_uniform_distribution(criteria)
    else:
        dist = criteria / norm
    return dist


# Inits
def _init_thresholds(init_threshold, stepsize, thresholds=None, tolerance=EPSILON):

    if thresholds is None:
        thresholds = np.arange(init_threshold, -1 - stepsize, -stepsize)

        # Otherwise rounding errors in feature importances fuck your shit up.
        thresholds[0] = thresholds[0] - tolerance

    return thresholds


def _init_m_avl(m_codes, m_avl=None):
    if m_avl is None:
        return np.arange(m_codes.shape[0], dtype=int)
    else:
        return m_avl


def _init_a_src_a_tgt(q_code=None, a_src=None, a_tgt=None):
    if q_code is not None:
        a_src, a_tgt, _ = code_to_query(q_code)
        return a_src, a_tgt
    else:
        return a_src, a_tgt




import numpy as np

# Helpers - Data Handling
def _dummy_array(nb_rows):
    a = np.empty((nb_rows, 1))
    a.fill(np.nan)
    return a


def _pad_proba(classes, all_classes):
    idx = _map_classes(classes, all_classes)

    def pad(X):
        R = np.zeros((X.shape[0], len(all_classes)))
        R[:, idx] = X
        return R

    return pad


def _map_classes(classes, all_classes):
    sorted_idx = np.argsort(all_classes)
    matches = np.searchsorted(all_classes[sorted_idx], classes)
    return sorted_idx[matches]


def _select_numeric(idx):
    def select(X):
        if X.ndim == 2:
            return X.take(idx, axis=1)
        else:
            return X

    return select


def _select_nominal(idx):
    def select(X):
        if isinstance(X, list):
            return X[idx]
        elif isinstance(X, np.ndarray):
            return X

    return select


import numpy as np
from functools import partial
from sklearn.preprocessing import maxabs_scale, minmax_scale
from sklearn.metrics import f1_score, r2_score, mean_squared_error, accuracy_score
from sklearn.model_selection import train_test_split
from ..utils.inference_tools import _dummy_array
from ..utils import DESC_ENCODING, TARG_ENCODING, MISS_ENCODING, get_i_o


def dummy_evaluation(
    data, m_codes, m_list, i_list, random_state=42, test_size=0.2, **kwargs
):
    return _dummy_evaluation(m_codes)


def base_evaluation(
    data,
    m_codes,
    m_list,
    i_list,
    random_state=42,
    test_size=0.2,
    per_attribute_normalization=False,
    consider_imputations=False,
    **kwargs
):

    # Data
    _, X = train_test_split(data, test_size=test_size, random_state=random_state)

    m_score = _model_evaluation(X, m_list, m_codes)

    if consider_imputations:
        i_score = _imputer_evaluation(X, i_list)
        m_score = normalize_m_score_relative_to_imputation(
            m_score, i_score, per_attribute_normalization=per_attribute_normalization
        )
    else:
        dummy_evaluation = _dummy_evaluation(m_codes)
        m_score = normalize_m_score(
            m_score,
            dummy_evaluation,
            per_attribute_normalization=per_attribute_normalization,
        )

    return m_score


def normalize_m_score_relative_to_imputation(
    m_score, i_score, per_attribute_normalization=False
):
    diff = m_score - i_score

    # All the models that do worse than the baseline
    diff[np.where(diff < 0)] = 0

    # Per attribute relative scaling.
    if per_attribute_normalization:
        norm_m_score = maxabs_scale(diff, axis=0)
    else:
        norm_m_score = maxabs_scale(diff.flat, axis=0).reshape(diff.shape)

    assert np.min(norm_m_score) >= 0
    return norm_m_score


def normalize_m_score(m_score, dummy_evaluation, per_attribute_normalization=False):
    # Per attribute relative scaling.
    if per_attribute_normalization:
        norm_m_score = minmax_scale(m_score, axis=0)
    else:
        norm_m_score = minmax_scale(m_score.flat, axis=0).reshape(m_score.shape)

    # Set zeroes again
    norm_m_score[np.where(dummy_evaluation != 1)] = 0.0
    norm_m_score[np.where(dummy_evaluation == 1)] = (
        norm_m_score[np.where(dummy_evaluation == 1)]
        + (1 - norm_m_score[np.where(dummy_evaluation == 1)]) * 0.01
    )

    assert np.min(norm_m_score) >= 0
    return norm_m_score


def _model_evaluation(X, m_list, m_codes):
    m_score = np.zeros(m_codes.shape)
    for m_idx, mod in enumerate(m_list):
        i, o = get_i_o(X, mod.desc_ids, mod.targ_ids, filter_nan=True)

        # i, o = X[:, mod.desc_ids], X[:, mod.targ_ids]

        multi_target = o.shape[1] != 1
        if not multi_target:
            # If output is single variable, we need 1D matrix
            o = o.ravel()

        y_true = o
        y_pred = mod.predict(i)

        metric = _select_metric(mod)
        mod.score = _calc_performance(y_true, y_pred, metric, multi_target)
        m_score[m_idx, mod.targ_ids] = mod.score

        # m_score[m_idx, mod.targ_ids] = (mod.score - i_score[mod.targ_ids])/mod.score
    return m_score


def _imputer_evaluation(X, i_list):
    # Evaluate imputers
    i_score = np.zeros(len(i_list))
    for i_idx, imp in enumerate(i_list):
        _, o = get_i_o(X, [], [i_idx], filter_nan=True)

        # o = X[:, i_idx]

        y_true = o
        y_pred = imp.transform(_dummy_array(len(o))).ravel()

        metric = _select_metric(imp)
        imp.score = _calc_performance(y_true, y_pred, metric)
        i_score[i_idx] = imp.score
    return i_score


def _dummy_evaluation(m_codes):
    m_score = np.zeros(m_codes.shape)
    m_score[np.where(m_codes == TARG_ENCODING)] = 1
    return m_score


def _select_metric(model):
    if model.out_kind in {"nominal"}:
        metric = accuracy_score
        #metric = partial(f1_score, average="macro")
    else:
        metric = normalized_root_mean_squared_error
        # metric = r2_score
    return metric


def _calc_performance(y_true, y_pred, metric, multi_target=False):

    if multi_target:
        performance = [
            metric(y_true[:, i], y_pred[:, i]) for i in range(y_true.shape[1])
        ]
    else:
        performance = metric(y_true, y_pred)
    return performance


def normalized_root_mean_squared_error(y_true, y_pred):
    """Compute normalized mean squared error
    Cf. missForest
    
    Returns:
        NMRSE -- Result
    """
    mse = mean_squared_error(y_true, y_pred)
    var = np.var(y_true)

    nrmse = np.sqrt(mse / var)
    return 1 - nrmse



# -*- coding: utf-8 -*-
"""
    Setup file for mercs.
    Use setup.cfg to configure your project.
"""
import sys

from pkg_resources import VersionConflict, require
from setuptools import setup

try:
    require('setuptools>=38.3')
except VersionConflict:
    print("Error: version of setuptools is too old (<38.3)!")
    sys.exit(1)


if __name__ == "__main__":
    setup()



import networkx as nx
import pydot
import tempfile
import warnings
from shutil import copyfile

from IPython.display import display, Image, SVG


def diagram_to_dotstring(g, fi_labels=False, ortho=False, rankdir="BT"):
    # Layout
    if fi_labels:
        for e in g.edges():
            g.edges()[e]["label"] = "{0:.2f}".format(g.edges()[e].get("fi", 0))

    dot = nx.drawing.nx_pydot.to_pydot(g)

    dot.set("rankdir", rankdir)

    if ortho:
        dot.set("splines", "ortho")

    dotstring = dot.to_string()

    return dotstring


def dotstring_to_image(dotstring):
    # Create temporary dotfile
    f_dotstring = tempfile.NamedTemporaryFile("w", delete=False, suffix=".dot")
    f_dotstring.write(dotstring)
    f_dotstring.close()

    # Read in PyDot
    (graph,) = pydot.graph_from_dot_file(f_dotstring.name)

    # Create temporary image file
    f_img = tempfile.NamedTemporaryFile("w", delete=False, suffix=".png")
    f_img.close()

    f_vec = tempfile.NamedTemporaryFile("w", delete=False, suffix=".svg")
    f_vec.close()

    (graph,) = pydot.graph_from_dot_file(f_dotstring.name)
    graph.write_png(f_img.name)
    graph.write_svg(f_vec.name)

    return f_img, f_vec


def show_diagram(diagram, kind="svg", fi=False, ortho=False, **kwargs):

    try:
        get_ipython()
        dotstring = diagram_to_dotstring(diagram, fi_labels=fi, ortho=ortho)

        f_img, f_vec = dotstring_to_image(dotstring)

        if kind in {"vector", "vec", "svg"}:
            return display(SVG(f_vec.name))
        elif kind in {"image", "img", "png"}:
            return display(Image(f_img.name, **kwargs))
        else:
            msg = """
            Did not recognize kind: {}
            Supported: ['vector', 'image']
            """.format(
                kind
            )
            raise NotImplementedError(msg)
    except BaseException as e:
        msg = """
        We can only show these diagrams in the ipython shell if you are calling this from the ipython shell.
        Exception was: {}
        """.format(e)
        warnings.warn(msg)


def save_diagram(diagram, fname, kind="svg", fi=False, ortho=False):
    dotstring = diagram_to_dotstring(diagram, fi_labels=fi, ortho=ortho)
    f_img, f_vec = dotstring_to_image(dotstring)

    if kind in {"vector", "vec", "svg"}:
        copyfile(f_vec.name, fname)
    elif kind in {"image", "img", "png"}:
        copyfile(f_img.name, fname)
    else:
        msg = """
        Did not recognize kind: {}
        Supported: ['vector', 'image']
        """.format(
            kind
        )
        raise NotImplementedError(msg)
    return



import numpy as np
from dask import delayed

from .compose import o, x
from ..graph.network import get_ids, node_label, get_nodes

from ..utils import debug_print

VERBOSITY = 0


class CompositeModel(object):
    def __init__(self, diagram, desc_ids=None, targ_ids=None):

        # Assign desc and targ ids
        if desc_ids is not None:
            self.desc_ids = desc_ids
        elif "desc_ids" in diagram.graph:
            self.desc_ids = diagram.graph["desc_ids"]
        else:
            self.desc_ids = get_ids(diagram, kind="desc")

        if targ_ids is not None:
            self.targ_ids = targ_ids
        elif "targ_ids" in diagram.graph:
            self.targ_ids = diagram.graph["targ_ids"]
        else:
            self.targ_ids = get_ids(diagram, kind="targ")

        self.desc_ids = list(self.desc_ids)
        self.targ_ids = list(self.targ_ids)

        self.feature_importances_ = self.extract_feature_importances(diagram)

        self.predict = _get_predict(diagram, self.targ_ids)
        self.out_kind = self.extract_out_kind(diagram, self.targ_ids)

        debug_print(self.out_kind)

        if self.out_kind in {"nominal", "mix"}:

            self.nominal_targ_ids = self.extract_nominal_targ_ids(
                diagram, self.targ_ids
            )

            self.predict_proba = _get_predict_proba(diagram, self.nominal_targ_ids)
            self.classes_ = _get_classes(diagram, self.nominal_targ_ids)

        return

    def extract_feature_importances(self, diagram, aggregation=np.sum):
        fi = []
        for idx in self.desc_ids:
            fi_idx = [
                d.get("fi", 0)
                for src, tgt, d in diagram.edges(data=True)
                if d.get("idx", 0) == idx
            ]
            fi.append(aggregation(fi_idx))

        norm = np.linalg.norm(fi, 1)
        fi = fi / norm
        return fi

    @staticmethod
    def extract_nominal_targ_ids(diagram, targ_ids):
        nominal_targ_ids = [
            t
            for t in targ_ids
            if diagram.node[node_label(t, kind="data")]["type"] == "nominal"
        ]
        return nominal_targ_ids

    @staticmethod
    def extract_out_kind(diagram, targ_ids):
        out_kinds = [diagram.node[node_label(t, kind="data")]["type"] for t in targ_ids]
        out_kinds = set(out_kinds)

        if len(out_kinds) == 1:
            return out_kinds.pop()
        else:
            return "mix"


def _get_classes(diagram, nominal_targ_ids):

    nominal_targ_ids.sort()

    # The prob and vote nodes have complete information on the classes
    targ_classes_ = [
        diagram.node[node_label(t, kind="prob")]["classes"] for t in nominal_targ_ids
    ]

    if len(targ_classes_) == 1:
        return targ_classes_.pop()
    else:
        return targ_classes_


def _get_predict_proba(diagram, nominal_targ_ids):

    nominal_targ_ids.sort()
    rel_nodes = [node_label(t, kind="prob") for t in nominal_targ_ids]
    tgt_methods = [diagram.nodes[n]["dask"] for n in rel_nodes]

    if len(tgt_methods) == 1:
        # We can safely return the only method
        return tgt_methods.pop()
    else:
        # We need to package them together
        collector = delayed(x)(*tgt_methods, return_type=list)
        return collector


def _get_predict(diagram, targ_ids):
    """
    Compose single predict function for a diagram.

    Parameters
    ----------
    diagram

    Returns
    -------

    """
    targ_ids.sort()

    rel_nodes = [node_label(t, kind="data") for t in targ_ids]
    tgt_methods = [diagram.nodes[n]["dask"] for n in rel_nodes]

    if len(tgt_methods) == 1:
        # We can safely return the only method
        return tgt_methods.pop()
    else:
        # We need to package them together
        collector = delayed(np.stack)(tgt_methods, axis=1)
        return collector


import numpy as np

def get_i_o(data, desc_ids, targ_ids, filter_nan=True):
    if filter_nan:
        i_o = data[:, desc_ids+targ_ids]

        mask = ~np.any(np.isnan(i_o), axis=1)
        i_o_filtered = i_o[mask]

        # This works because we ordered the columns very specifically.
        i = i_o_filtered[:, :len(desc_ids)]
        o = i_o_filtered[:, len(desc_ids):]

    else:
        i, o = data[:, desc_ids], data[:, targ_ids]
    return i, o